#include "client.hh"

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <iostream>
#include <cerrno>
#include <sstream>
#include <stdio.h>

using namespace std;
using namespace libwebots::communication;

Client *Client::s_instance = 0;

/**
 * @class libwebots::communication::Client
 * @brief Controller client communication class.
 *
 * The Client class is used in a Webots controller to send and receive messages
 * from the physics plugin. Multiple controllers each have their own client
 * instance and can communicate with the same physics plugin. A client connects
 * to the plugin server using a UNIX domain socket. The path of this socket
 * is constructed from the PPID (parent process id).
 *
 * @par Creating a client
 * You can get an instance for a Client in your controller using
 * Client::Instance. The first time you get a client you will need to connect
 * it to the server using Connect().
 *
 * @par Sending messages
 * You can send messages to your physics plugin modules using the Send() method.
 * The send method accepts a Message object. You can construct a Message object
 * manually (using Messages::Create), or you can use one of the builder helper
 * classes provided with the plugin modules in libwebots.
 *
 * @par Receiving messages
 * There are two ways to receive and process messages that are sent by the
 * physics plugin modules. The first is to use the Handlers::AddHandler() method.
 * You can use this method to register a callback function for a particular
 * message type. Note that Handlers is a base class of Client, so you can
 * simple call @p client.AddHandler. When using this method, you @b must call
 * ProcessMessages() in your webots main loop. If you do not, Webots will hang
 * because it cannot finish sending messages when the buffer gets full (unless
 * asynchronous mode is used). The second method is to explicitly handle
 * messages manually when they are received. You can use Receive() to receive
 * a new message in your Webots controller main loop. Finally, you can write
 * your own ProcessMessages() type of method (see ProcessMessages() for more
 * information).
 *
 * @par Example
 * @code
 * // Get the client singleton instance
 * webots::Supervisor controller;
 * libwebots::communication::Client &client(libwebots::communication::Client::Instance());
 *
 * // Here you can add some handlers to the client using AddHandler:
 * // client.AddHandler(object, messagetype, &Class::HandlerCallback);
 *
 * int timeStep = static_cast<int>(controller.getBasicTimeStep() * 1000);
 *
 * // Connect the client to the server (in the physics plugin)
 * client.Connect();
 *
 * while (controller.step(timeStep) != -1)
 * {
 *   // Make sure to call ProcessMessages() to receive messages from the plugin
 *   // and deliver them to any matching handlers you setup with AddHandler
 *   client.ProcessMessages();
 * }
 * @endcode
 */

/**
 * @brief Client singleton instance
 *
 * Get a reference to the client instance. Always use this to get a reference
 * to a client from your Webots controller. If you want to initialize this
 * singleton client in asynchronous mode, use Client::Initialize.
 *
 */
Client &
Client::Instance()
{
	if (!s_instance)
	{
		s_instance = new Client();
	}

	return *s_instance;
}

/**
 * @brief Initialize client singleton
 * @param async synchronous/asynchronous mode
 *
 * Initialize the client singleton in synchronous or asynchronous mode. By
 * default, the client singleton instance will be created in asynchronous mode.
 * You do not need to use this method (use Client::Instance instead, it
 * implicitly initializes the client singleton in synchronous mode), unless you
 * need a client in asynchronous mode.
 */
Client &
Client::Initialize(bool async)
{
	if (s_instance)
	{
		delete s_instance;
	}

	s_instance = new Client(async);
	return *s_instance;
}

/**
 * @brief Finalize the singleton instance
 *
 * Finalizes the singleton instance of the client.
 */
void
Client::Finalize()
{
	if (s_instance)
	{
		delete s_instance;
		s_instance = 0;
	}
}

/**
 * @brief Create a new client
 * @param async synchronous/asynchronous mode
 *
 * Create a new client object. You should never need to construct
 * a client yourself, always use the singleton Client::Instance. A client
 * can optionally be constructed to work in asynchronous mode. In this mode,
 * messages will be received and sent asynchronously. It is not guaranteed that
 * messages will be delivered at each step. The default mode is synchronous.
 */
Client::Client(bool async)
:
	d_state(State::Closed),
	d_async(async)
{
	if (d_async)
	{
		pthread_mutex_init (&d_rwLock, NULL);
	}
}

/**
 * @brief Create a new client for a file descriptor
 * @param fd the file descriptor
 * @param async synchronous/asynchronous mode
 *
 * Creates a new client for a specific file descriptor. Do not use this
 * constructor in your controller, it is for internal use.
 *
 */
Client::Client(int fd, bool async)
:
	Socket(fd),
	d_state(State::Closed),
	d_async(async)
{
	if (d_async)
	{
		pthread_mutex_init (&d_rwLock, NULL);
	}

	ConnectDone();
}

/**
 * @brief Connect the client to the server
 *
 * Connects the client to the server. After initializing the client, you need
 * to explicitly connect it to the server. After the method returns, you can
 * check whether the connection was successful using Connected()
 */
void
Client::Connect()
{
	if (Fd() == -1)
	{
		char const *tmp = getenv("TMPDIR");

		if (!tmp)
		{
			tmp = "/tmp";
		}

		stringstream templ;

		templ << tmp << "/webots-communication." << getppid();

		Open(templ.str());
	}

	if (Fd() == -1)
	{
		return;
	}

	if (Connecting() || Connected())
	{
		return;
	}

	if (d_async)
	{
		SetNonBlocking(true);
	}

	int ret = ::connect(Fd(), AddressInfo()->ai_addr, AddressInfo()->ai_addrlen);

	if (ret == -1)
	{
		if (errno == EINPROGRESS)
		{
			d_state = State::Connecting;
		}
		else
		{
			cerr << "[Client] Could not connect to communication server: " << strerror(errno) << endl;
			Close();
		}
	}
	else
	{
		cout << "[Client] Connected to communication server..." << endl;
		ConnectDone();
	}
}

void
Client::ConnectDone()
{
	d_state = State::Connected;

	// Start thread to read/write from/to the socket
	if (d_async)
	{
		SetNonBlocking(false);
		pthread_create (&d_rwThread, NULL, Client::ReadWriteThread, this);
	}
}

/**
 * @brief Close connection to the server.
 *
 * Close the connection to the server.
 *
 **/
void
Client::Close()
{
	Socket::Close();

	if (d_async)
	{
		pthread_mutex_lock (&d_rwLock);
		bool joinit = d_state == State::Connected || d_state == State::Closing;
		pthread_mutex_unlock (&d_rwLock);

		if (joinit)
		{
			pthread_join (d_rwThread, NULL);
		}
	}

	d_state = State::Closed;
}

bool
Client::FinishConnect()
{
	if (d_state != State::Connecting)
	{
		return true;
	}

	fd_set fdset;

	FD_ZERO(&fdset);
	FD_SET(Fd(), &fdset);

	struct timeval tv = {0, 0};

	if (::select(Fd() + 1, 0, &fdset, 0, &tv) != -1 && FD_ISSET(Fd(), &fdset))
	{
		ConnectDone();
		return true;
	}
	else
	{
		return false;
	}
}

void *
Client::ReadWriteThread(void *cl)
{
	Client *client = (Client *)cl;
	int fd = client->Fd();


	while (true)
	{
		char buffer[4096];

		ssize_t ret = ::recv(fd, buffer, sizeof(buffer), 0);

		if (ret > 0)
		{
			pthread_mutex_lock (&client->d_rwLock);
			client->d_data.append(buffer, (std::string::size_type)ret);
			pthread_mutex_unlock (&client->d_rwLock);
		}
		else if (ret <= 0)
		{
			pthread_mutex_lock (&client->d_rwLock);
			client->d_state = State::Closing;
			pthread_mutex_unlock (&client->d_rwLock);
			break;
		}
	}

	return 0;
}

/**
 * @brief Send message to physics plugin.
 * @param message the message to send
 *
 * Send a single message to the physics plugin. See the Message class for more
 * information how messages are constructed.
 *
 * @returns true if the message was sent, false otherwise
 *
 **/
bool
Client::Send(Message const &message)
{
	if (Closing())
	{
		Close();
	}

	if (Fd() == -1)
	{
		return false;
	}

	if (Connecting())
	{
		if (!FinishConnect())
		{
			return false;
		}
	}

	if (!Connected())
	{
		return false;
	}

	string serialized;

	if (!message.SerializeToString(&serialized))
	{
		return false;
	}

	stringstream s;

	size_t len = serialized.length();
	s << len << " " << serialized;

	serialized = s.str();

	return ::send(Fd(), serialized.c_str(), serialized.length(), 0) != 0;
}

/**
 * @brief Receive single message from physics plugin.
 * @param message message return value by reference
 *
 * Receive a single message (if possible) from the physics plugin.
 *
 * @returns true if a new message was received (available in message),
 *          or false otherwise.
 */
bool
Client::Receive(Message &message)
{
	size_t num;

	if (Closing())
	{
		Close();
		return false;
	}

	if (Closed())
	{
		return false;
	}

	std::stringstream s;

	if (d_async)
	{
		pthread_mutex_lock (&d_rwLock);
	}
	else
	{
		fd_set readfds;

		FD_ZERO(&readfds);
		FD_SET(Fd(), &readfds);

		timeval timeout = {0, 0};

		while (select(Fd() + 1, &readfds, 0, 0, &timeout) != -1 && FD_ISSET(Fd(), &readfds))
		{
			char buffer[4096];

			ssize_t ret = ::recv(Fd(), buffer, sizeof(buffer), 0);

			if (ret > 0)
			{
				d_data.append(buffer, (std::string::size_type)ret);
			}
			else if (ret == 0)
			{
				Close();
				break;
			}
		}
	}

	s.str(d_data);

	if (!(s >> num))
	{
		if (d_async)
		{
			pthread_mutex_unlock (&d_rwLock);
		}

		return false;
	}

	if (!s.ignore(1, ' '))
	{
		if (d_async)
		{
			pthread_mutex_unlock (&d_rwLock);
		}

		return false;
	}

	char *buffer = new char[num + 1];

	if (!s.read(buffer, (std::streamsize)num))
	{
		if (d_async)
		{
			pthread_mutex_unlock (&d_rwLock);
		}

		delete[] buffer;
		return false;
	}

	buffer[num] = '\0';

	try
	{
		message.ParseFromArray(buffer, num);
	}
	catch (std::exception e)
	{
		if (d_async)
		{
			pthread_mutex_unlock (&d_rwLock);
		}

		delete[] buffer;
		return false;
	}

	delete[] buffer;

	d_data = d_data.substr((std::string::size_type)s.tellg());

	if (d_async)
	{
		pthread_mutex_unlock (&d_rwLock);
	}

	return true;
}

/**
 * @brief Process messages from the physics plugin
 *
 * Call this method in your Webots controller main loop when you use message
 * handlers (Handlers::AddHandler()) to receive messages.
 *
 * @note This method will simply call Receive() and pass received messages to
 * Handlers::Call which will invoke registered callback handlers. You can
 * implement your own message processing, combining both callbacks and/or other
 * custom message handling routines if needed.
 *
 * @returns true if any messages were handled, false otherwise
 *
 **/
bool
Client::ProcessMessages()
{
	Message message;
	bool received = false;

	while (Receive(message))
	{
		Call(message);
		received = true;
	}

	return received;
}

bool
Client::IsState(State::Values state)
{
	if (d_async)
	{
		pthread_mutex_lock (&d_rwLock);
	}

	State::Values ret = d_state;

	if (d_async)
	{
		pthread_mutex_unlock (&d_rwLock);
	}

	return ret == state;
}

/**
 * @brief Check whether client is connecting
 *
 * Check whether the client is connecting to the server.
 *
 * @returns true if the client is connecting, false otherwise
 */
bool
Client::Connecting()
{
	return IsState(State::Connecting);
}

/**
 * @brief Check whether client is connected
 *
 * Check whether the client is connected to the server.
 *
 * @returns true if the client is connected, false otherwise
 */
bool
Client::Connected()
{
	return IsState(State::Connected);
}

/**
 * @brief Check whether client is closing the connection
 *
 * Check whether the client is closing the connection to the server.
 *
 * @returns true if the client is closing the connection, false otherwise
 */
bool
Client::Closing()
{
	return IsState(State::Closing);
}

/**
 * @brief Check whether client is disconnected
 *
 * Check whether the client is disconnected from the server.
 *
 * @returns true if the client is disconnected, false otherwise
 */
bool
Client::Closed()
{
	return IsState(State::Closed);
}
