#ifndef __LIBWEBOTS_COMMUNICATION_HANDLER_H__
#define __LIBWEBOTS_COMMUNICATION_HANDLER_H__

#include <libwebots/communication/message.pb.h>

namespace libwebots
{
namespace communication
{
	class Handler
	{
		int d_id;
		std::string d_match;

		public:
			Handler(int id, std::string const &match = "");
			virtual ~Handler();
			virtual void Emit(communication::Message const &message) = 0;

			int Id() const;
			std::string const &Match() const;
	};

	typedef ::google::protobuf::internal::FieldType FieldType;

	/**
	 * @brief Callback container class
	 *
	 * Class used to encapsulate an object member callback for a particular
	 * message type.
	 *
	 */
	template <typename Object, typename MsgType, FieldType field_type, bool is_packed>
	class HandlerCallback : public Handler
	{
		public:
			typedef MsgType Type;
			typedef ::google::protobuf::internal::MessageTypeTraits<Type> TypeTraits;

			typedef void (Object::*Callback)(Type const &message);
			typedef const ::google::protobuf::internal::ExtensionIdentifier<communication::Message,
			                                                                TypeTraits,
			                                                                field_type,
			                                                                is_packed> MessageType;


			HandlerCallback(Object &obj, MessageType &type, Callback callback, std::string const &match = "");

			virtual void Emit(communication::Message const &message);
		private:
			Object &d_obj;
			MessageType &d_message;
			Callback d_callback;
	};

	/**
	 * @brief Create new HandlerCallback
	 * @param obj the object
	 * @param type the message type
	 * @param callback the member callback function
	 * @param match the filter
	 *
	 * Create a new handler callback for a particular callback and message
	 * type
	 *
	 **/
	template <typename Object, typename Type, FieldType field_type, bool is_packed>
	HandlerCallback<Object, Type, field_type, is_packed>::HandlerCallback(Object &obj, MessageType &type, Callback callback, std::string const &match)
	:
		Handler(type.number(), match),
		d_obj(obj),
		d_message(type),
		d_callback(callback)
	{
	}

	/**
	 * @brief Call callback for particular message
	 * @message the message container
	 *
	 * Calls the callback contained in the handler for the particular
	 * extension this callback was registered for. The extension needs
	 * to exist in the provided @em message.
	 *
	 */
	template <typename Object, typename Type, FieldType field_type, bool is_packed>
	void HandlerCallback<Object, Type, field_type, is_packed>::Emit(communication::Message const &message)
	{
		(d_obj.*d_callback)(message.GetExtension(d_message));
	}
}
}

#endif /* __LIBWEBOTS_COMMUNICATION_HANDLER_H__ */

