#include "handlers.hh"

using namespace libwebots::communication;
using namespace std;

/**
 * @class libwebots::communication::Handlers
 * @brief Callback handler class
 *
 * This class is used to register callback handlers for particular messages.
 * Both Client and Server inherit from this class (i.e. both client and server
 * side use the same methods for receiving messages).
 *
 * The most important method of this class is the AddHandler() method. Note that
 * due to the extensive use of templates, this methods signature looks
 * particularly messy. It is however very easy and intuitive to use. See
 * AddHandler() documentation for more information.
 *
 **/

/**
 * @brief Call message handlers.
 * @param message the message
 *
 * Call all method handlers currently registered for the given message.
 *
 * @returns true if the message was handled by a callback, false otherwise
 */
bool
Handlers::Call(Message const &message)
{
	::google::protobuf::Reflection const &refl = *message.GetReflection();

	bool ret = false;

	for (vector<Handler *>::iterator iter = d_handlers.begin(); iter != d_handlers.end(); ++iter)
	{
		Handler &handler = **iter;

		::google::protobuf::FieldDescriptor const *desc = refl.FindKnownExtensionByNumber(handler.Id());

		if (desc && refl.HasField(message, desc) && (handler.Match() == "" || (message.has_match() && handler.Match() == message.match())))
		{
			handler.Emit(message);
			ret = true;
		}
	}

	return ret;
}

Handlers::~Handlers()
{
	for (vector<Handler *>::iterator iter = d_handlers.begin(); iter != d_handlers.end(); ++iter)
	{
		delete *iter;
	}
}
