#ifndef __LIBWEBOTS_COMMUNICATION_HANDLERS_H__
#define __LIBWEBOTS_COMMUNICATION_HANDLERS_H__

#include <libwebots/communication/handler.hh>
#include <vector>

namespace libwebots
{
namespace communication
{
	class Handlers
	{
		std::vector<Handler *> d_handlers;

		public:
			typedef ::google::protobuf::internal::FieldType FieldType;

			template <typename Object,
			          typename Type,
			          FieldType field_type,
			          bool is_packed>
			void AddHandler(Object &object,
			                const ::google::protobuf::internal::ExtensionIdentifier<Message,
			                        ::google::protobuf::internal::MessageTypeTraits<Type>,
			                        field_type,
			                        is_packed> &type,
			                typename HandlerCallback<Object, Type, field_type, is_packed>::Callback callback,
			                std::string const &match = "");

			bool Call(Message const &message);

			virtual ~Handlers();
	};

	/**
	 * @brief Register message callback handler
	 * @param object the callback object
	 * @param type the message type
	 * @param callback the callback member
	 * @param match a message match
	 *
	 * This method registers a message callback handler (see Client class).
	 * Although the signature of this method looks scary, it is very simple
	 * to use.
	 *
	 * The @em object parameter is a reference to an object on which to call
	 * the callback method. The @em type parameter indicates the message type
	 * to register the callback method for. The type of this parameter is
	 * a google protobuf extension type (see example below). The @em callback
	 * parameter is a C++ member function pointer, and can simply be
	 * specified as @p &Class::Member. The signature of this method should
	 * be:
	 *
	 * @code
	 * // MyMessage is a particular extension message
	 * void Member(MyMessage const &message);
	 * @endcode
	 *
	 * The last parameter (@em match) can be used to add a filter for
	 * particular messages that support filtering. Only the messages
	 * matching the filter will be received on this handler. One example
	 * of such a filter are the various Monitor messages
	 * (constraint, state). These usually include an optional filter on
	 * the name of the body which you monitor. When not supplying any filter,
	 * all messages (regardless of any filtering) will be delivered to
	 * the callback.
	 *
	 * @par Example
	 * @code
	 * // This goes in a .proto file:
	 * package libwebots.plugin.modules.messages;
	 *
	 * message MyMessage
	 * {
	 *   required string msg = 1;
	 * }
	 *
	 * extend libwebots.communication.Message
	 * {
	 *   optional MyMessage mymessage = 1001;
	 * }
	 *
	 * // This goes in a C++ file:
	 * class MyClass
	 * {
	 *   public:
	 *     void Handler(libwebots::plugin::modules::messages::MyMessage const &message)
	 *     {
	 *     }
	 * };
	 *
	 * MyClass obj;
	 *
	 * AddHandler(obj, libwebots::plugin::modules::messages::mymessage, &MyClass::Handler);
	 * @endcode
	 **/
	template <typename Object,
	          typename Type,
	          Handlers::FieldType field_type, bool is_packed>
	void Handlers::AddHandler(Object &object,
	                          const ::google::protobuf::internal::ExtensionIdentifier<Message,
	                                  ::google::protobuf::internal::MessageTypeTraits<Type>,
	                                  field_type,
	                                  is_packed> &type,
	                          typename HandlerCallback<Object, Type, field_type, is_packed>::Callback callback,
	                          std::string const &match)
	{
		d_handlers.push_back(new HandlerCallback<Object, Type, field_type, is_packed>(object,
		                                                                              type,
		                                                                              callback,
		                                                                              match));
	}
}
}

#endif /* __LIBWEBOTS_COMMUNICATION_HANDLERS_H__ */

