#ifndef __LIBWEBOTS_COMMUNICATION_MESSAGES_H__
#define __LIBWEBOTS_COMMUNICATION_MESSAGES_H__

#include <libwebots/communication/message.pb.h>

namespace libwebots
{
namespace communication
{
	/**
	 * @brief Message factory
	 *
	 * Convenience class for wrapping extension messages in the base
	 * prototype message. Use Messages::Create() to wrap a particular extension
	 * message into a Message class which can be sent using a Client.
	 *
	 * @par Example
	 * @code
	 * // This goes in a .proto file:
	 * package libwebots.plugin.modules.messages;
	 *
	 * message MyMessage
	 * {
	 *   required string msg = 1;
	 * }
	 *
	 * extend libwebots.communication.Message
	 * {
	 *   optional MyMessage mymessage = 1001;
	 * }
	 *
	 * // This goes in your controller
	 * Client &client(Client::Instance());
	 *
	 * MyMessage message;
	 * message.set_msg("Hello World!");
	 *
	 * client.Send(Messages::Create(mymessage, message));
	 * @endcode
	 *
	 **/
	class Messages
	{
		typedef ::google::protobuf::internal::FieldType FieldType;

		public:
			/**
			 * @brief Wrap a message extension.
			 * @param type extension type
			 * @param value the extension message
			 * @param match optional match to set in message
			 *
			 * Wrap an extension message into the base Message
			 * class. This will set the extension message
			 * (@em value) of the given @em type in a new Message
			 * class. The optional @em match will also be set in
			 * the returned message.
			 *
			 * @returns a new Message
			 *
			 **/
			template <typename Type,
			          FieldType field_type,
			          bool is_packed>
			static communication::Message Create(const ::google::protobuf::internal::ExtensionIdentifier<communication::Message,
			::google::protobuf::internal::MessageTypeTraits<Type>, field_type, is_packed> &type,
			typename ::google::protobuf::internal::MessageTypeTraits<Type>::ConstType value, std::string const &match = "")
			{
				communication::Message message;

				if (match != "")
				{
					message.set_match(match);
				}

				*message.MutableExtension(type) = value;

				return message;
			}
	};
}
}

#endif /* __LIBWEBOTS_COMMUNICATION_MESSAGES_H__ */

