#include "server.hh"
#include <cerrno>
#include <cstring>
#include <iostream>
#include <cstdlib>
#include <sys/un.h>

using namespace std;
using namespace libwebots::communication;

Server *Server::s_instance = 0;

/**
 * @class libwebots::communication::Server
 * @brief Plugin server communication class
 *
 * The server class is used in the physics plugin. Multiple Client objects can
 * connect and communicate to the server. The server and client communicate
 * using a UNIX domain socket. You should never need to use this class yourself
 * as it is managed by libwebots::plugin::Plugin.
 *
 **/

/**
 * @brief Create new server
 * @param async true if asynchronous mode, false otherwise (default)
 *
 * Create a new server in synchronous or asynchronous mode. This will
 * automatically generate a unix socket file name based on the PID of the
 * current process. It will open a socket on this file and then starts
 * listening for incoming client connections.
 *
 **/
Server::Server(bool async)
:
	d_currentClient(0),
	d_async(async)
{
	d_filename = GenerateSocketFileName();

	// Generate a new filename to listen on
	if (d_filename == "")
	{
		cerr << "[Server] Could not create temporary file for communication: " << strerror(errno) << endl;
		return;
	}

	Open(d_filename);

	if (Fd() == -1)
	{
		cerr << "[Server] Could not create communication socket: " << strerror(errno) << endl;
		Close();

		return;
	}

	// Set to reuse address, to be sure
	int one = 1;
	setsockopt(Fd(), SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

	SetNonBlocking();

	// Find the socket
	if (::bind(Fd(), AddressInfo()->ai_addr, AddressInfo()->ai_addrlen) != 0)
	{
		cerr << "[Server] Could not bind communication socket: " << strerror(errno) << endl;
		Close();

		return;
	}

	if (::listen(Fd(), 100) != 0)
	{
		cerr << "[Server] Could not listen on communication socket: " << strerror(errno) << endl;
		Close();

		return;
	}
}

/**
 * @brief Close server
 *
 * Close the server. Connected clients will be disconnected.
 *
 **/
void
Server::Close()
{
	vector<Client *>::iterator iter;

	for (iter = d_clients.begin(); iter != d_clients.end(); ++iter)
	{
		delete *iter;
	}

	d_currentClient = 0;

	d_clients.clear();

	Socket::Close();

	if (d_filename != "")
	{
		::unlink(d_filename.c_str());
	}
}

Server::~Server()
{
	Close();
}

/**
 * @brief Generate the socket filename
 *
 * Generate the unix socket domain file name. This file name is located in
 * @p TMPDIR @p / @p webots-communication.PID where PID is the current process
 * id. This convention allows a single unique socket per webots process and
 * allows controllers to automatically connect (they construct the file path
 * using getppid() since there parent is the Webots process). This is
 * considered neato.
 *
 * @returns the generated filename
 *
 **/
string
Server::GenerateSocketFileName()
{
	char const *tmp = getenv("TMPDIR");

	if (!tmp)
	{
		tmp = "/tmp";
	}

	stringstream templ;

	templ << tmp << "/webots-communication." << getpid();

	return templ.str();
}

/**
 * @brief Initialize server singleton
 * @param async asynchronous mode
 *
 * Initializes the server singleton
 *
 * @returns the server singleton instance
 *
 **/
Server &
Server::Initialize(bool async)
{
	if (!s_instance)
	{
		s_instance = new Server(async);
	}

	return *s_instance;
}

/**
 * @brief Get server singleton instance
 *
 * Get the server singleton instance
 *
 * @returns the server singleton instance
 *
 **/
Server &
Server::Instance()
{
	return Server::Initialize();
}

/**
 * @brief Finalize server singleton instance
 *
 * Finalize the server singleton instance
 *
 **/
void
Server::Finalize()
{
	if (s_instance)
	{
		delete s_instance;
		s_instance = 0;
	}
}

/**
 * @brief Send message to clients
 * @param message a message
 *
 * Send a message to all connected clients.
 *
 **/
bool
Server::Send(Message const &message)
{
	vector<Client *>::iterator iter;
	bool ret = true;

	for (iter = d_clients.begin(); iter != d_clients.end(); ++iter)
	{
		ret &= (*iter)->Send(message);
	}

	return ret;
}

/**
 * @brief Accept incoming connections
 *
 * Accept pending incoming connections
 *
 **/
void
Server::Accept()
{
	if (Fd() == -1)
	{
		return;
	}

	fd_set fdset;

	FD_ZERO(&fdset);
	FD_SET(Fd(), &fdset);

	struct timeval tv = {0, 0};

	while (select(Fd() + 1, &fdset, 0, 0, &tv) > 0)
	{
		struct sockaddr_un addr;
		socklen_t addrlen = sizeof(addr);
		struct sockaddr *ad = reinterpret_cast<struct sockaddr *>(&addr);

		int fd = ::accept(Fd(), ad, &addrlen);

		if (fd != -1)
		{
			Client *client = new Client(fd, d_async);
			d_clients.push_back(client);
		}
		else
		{
			cerr << "[Server] Failed to accept new client: " << strerror(errno) << endl;
		}
	}
}

/**
 * @brief Receive incoming messages
 * @param client return value for client
 * @param message return value for message
 *
 * Receive a single incoming message from any client. If a message was received,
 * the client is set to the client from which the message was received.
 *
 * @returns true if a message was received, false otherwise
 *
 **/
bool
Server::Receive(Client *&client, Message &message)
{
	vector<Client *>::iterator iter;

	bool ret = false;

	iter = d_clients.begin();

	while (iter != d_clients.end())
	{
		if ((*iter)->Receive(message))
		{
			d_currentClient = *iter;

			if (!Call(message))
			{
				client = *iter;
				ret = true;
				break;
			}

			d_currentClient = 0;
		}
		else if (!(*iter))
		{
			delete *iter;
			iter = d_clients.erase(iter);
		}
		else
		{
			++iter;
		}
	}

	return ret;
}
