#include "socket.hh"
#include <cstdlib>
#include <sys/un.h>
#include <fcntl.h>
#include <cerrno>
#include <iostream>

using namespace std;
using namespace libwebots::communication;

/**
 * @class libwebots::communication::Socket
 * @brief Basic socket class
 *
 * Basic socket class for communication over UNIX domain sockets.
 *
 **/

/**
 * @brief Create a new socket
 *
 * Create a new empty socket
 *
 **/
Socket::Socket()
:
	d_fd(-1),
	d_info(0)
{
}

/**
 * @brief Create a new wrapper for an existing socket
 * @param fd a file descriptor
 *
 * Create a new wrapper Socket for an existing socket (file descriptor)
 *
 **/
Socket::Socket(int fd)
:
	d_fd(fd),
	d_info(0)
{
}

Socket::~Socket()
{
	Close();

	if (d_info)
	{
		free(d_info->ai_addr);
		free(d_info);
	}
}

/**
 * @brief Create and open a new unix domain socket
 * @param filename the unix domain socket path
 *
 * Create a new socket and Open() a unix domain socket.
 *
 **/
Socket::Socket(string const &filename)
:
	d_fd(-1),
	d_info(0)
{
	Open(filename);
}

/**
 * @brief Assign file descriptor to socket
 * @param fd a file descriptor
 *
 * Assign a file descriptor to the socket object. This only works for empty
 * socket objects.
 *
 **/
void
Socket::Open(int fd)
{
	if (d_fd == -1)
	{
		d_fd = fd;
	}
}

/**
 * @brief Open unix domain socket
 * @param filename unix domain socket path
 *
 * Open a unix domain socket on a particular path.
 *
 **/
void
Socket::Open(string const &filename)
{
	if (filename == "" || d_fd != -1)
	{
		return;
	}

	d_info = CreateAddressInfo(filename);
	d_fd = ::socket(d_info->ai_family, d_info->ai_socktype, d_info->ai_protocol);
}

/**
 * @brief Get the socket address info
 *
 * Get the address info associated to the socket
 *
 * @returns the address info
 *
 **/
struct addrinfo const *
Socket::AddressInfo() const
{
	return d_info;
}

/**
 * @brief Close the socket
 *
 * Close the socket
 *
 **/
void
Socket::Close()
{
	if (d_fd != -1)
	{
		shutdown(d_fd, SHUT_RDWR);
		close(d_fd);

		d_fd = -1;
	}
}

/**
 * @brief Get the file descriptor
 *
 * Get the socket file descriptor
 *
 * @returns the socket file descriptor
 *
 **/
int
Socket::Fd() const
{
	return d_fd;
}

struct addrinfo *
Socket::CreateAddressInfo(string const &filename)
{
	struct sockaddr_un *addr = new sockaddr_un();

	addr->sun_family = AF_UNIX;
	strncpy(addr->sun_path, filename.c_str(), 107);
	addr->sun_path[107] = '\0';

	struct addrinfo *info = static_cast<addrinfo *>(calloc(1, sizeof(addrinfo)));

	info->ai_family = AF_UNIX;
	info->ai_socktype = SOCK_STREAM;
	info->ai_addrlen = sizeof(struct sockaddr_un);
	info->ai_addr = (sockaddr *)addr;

	return info;
}

/**
 * @brief Validate socket
 *
 * Get whether the socket is assigned
 *
 * @returns whether the socket is assigned
 *
 **/
Socket::operator bool() const
{
	return d_fd != -1;
}

/**
 * @brief Set nonblocking/blocking socket
 * @param val true if nonblocking, false if blocking
 *
 * Set or unset the socket non-blocking mode
 *
 **/
void
Socket::SetNonBlocking(bool val)
{
	// Set non blocking
	int flags = fcntl(Fd(), F_GETFL, 0);

	if (flags == -1)
	{
		cerr << "Could not set socket non-blocking: " << strerror(errno) << endl;
		return;
	}

	if (val)
	{
		fcntl(Fd(), F_SETFL, flags | O_NONBLOCK);
	}
	else
	{
		fcntl(Fd(), F_SETFL, flags & ~O_NONBLOCK);
	}
}
