#include "body.hh"

#include <Eigen/SVD>
#include <Eigen/LU>

using namespace libwebots::plugin;
using namespace std;

/**
 * @class libwebots::plugin::Body
 * @brief Wrapper class for ODE body
 *
 * This class provides a convenient wrapper for an ODE body. The class inherits
 * from a map of joint ids to child bodies (i.e. the children of this body
 * connected through joints). It provides convenient accessors for translation,
 * rotation, angular velocity, linear velocity etc. using the Eigen library.
 *
 **/

/**
 * @brief Create a new body
 * @param def a Webots def
 *
 * Wrap a body from a Webots def
 *
 */
Body::Body(std::string const &def)
:
	d_id(dWebotsGetBodyFromDEF(def.c_str())),
	d_iterated(false),
	d_def(def)
{
	LocalTransform();
}

/**
 * @brief Wrap an ODE body
 * @param id an ODE body
 *
 * Wrap an ODE body
 *
 **/
Body::Body(dBodyID id)
:
	d_id(id),
	d_iterated(false)
{
	LocalTransform();
}

/**
 * @brief Copy constructor
 *
 * Copy constructor
 *
 **/
Body::Body(Body const &other)
:
	std::map<dJointID,Body>(),
	d_id(other.d_id),
	d_iterated(false),
	d_def(other.d_def),
	d_localRotation(other.d_localRotation),
	d_localTranslation(other.d_localTranslation)

{
}

/**
 * @brief Assignment
 *
 * Assignment
 *
 * @returns reference to the object
 *
 **/
Body &
Body::operator=(Body const &other)
{
	if (this == &other)
	{
		return *this;
	}

	d_id = other.d_id;
	d_iterated = false;

	d_localRotation = other.d_localRotation;
	d_localTranslation = other.d_localTranslation;
	d_def = other.d_def;

	clear();

	FindChildren();

	return *this;
}

/**
 * @brief Validate body
 *
 * Validate body
 *
 * @returns true if body is valid, false otherwise
 *
 */
Body::operator bool() const
{
	return d_id != 0;
}

void
Body::LocalTransform()
{
	d_localTranslation = Eigen::Vector3d::Zero();
	d_localRotation = Eigen::Matrix3d::Identity();

	if (!d_id)
	{
		return;
	}

	int num = dBodyGetNumJoints(d_id);

	for (int i = 0; i < num; ++i)
	{
		dJointID joint = dBodyGetJoint(d_id, i);
		dJointType type = dJointGetType(joint);

		dBodyID b1 = dJointGetBody(joint, 0);
		dBodyID b2 = dJointGetBody(joint, 1);

		if (b2 == d_id)
		{
			Body parent(b1);

			d_localRotation = Rotation() * parent.Rotation().inverse().transpose();
			Eigen::Vector3d parentTranslation;

			if (type == dJointTypeHinge)
			{
				// Setup local transformation with respect to the anchor
				dVector3 odeanchor;

				dJointGetHingeAnchor(joint, odeanchor);
				parentTranslation = Eigen::Vector3d(odeanchor);

			}
			else
			{
				parentTranslation = parent.Translation();
			}

			d_localTranslation = Translation() - parentTranslation;
			break;
		}
	}
}

void
Body::FindChildren()
{
	if (!d_id || d_iterated)
	{
		return;
	}

	int num = dBodyGetNumJoints(d_id);
	d_iterated = true;

	for (int i = 0; i < num; ++i)
	{
		dJointID joint = dBodyGetJoint(d_id, i);
//		dJointType type = dJointGetType(joint);

		dBodyID b1 = dJointGetBody(joint, 0);
		dBodyID b2 = dJointGetBody(joint, 1);

		if (b1 == d_id && b2)
		{
			(*this)[joint] = Body(b2);
		}
	}
}

/**
 * @brief Get the ODE body id
 *
 * Get the wrapped ODE body id
 *
 * @returns the wrapped ODE body id
 *
 */
dBodyID
Body::Id() const
{
	return d_id;
}

/**
 * @brief Get whether a body is a descendent of this body
 *
 * Get whether the given body is a descendent of this body
 *
 * @returns true if id is a descendent of this body
 *
 **/
bool
Body::Includes(dBodyID id) const
{
	Body::const_iterator iter;

	if (d_id == 0 || id == 0)
	{
		return false;
	}

	if (id == d_id)
	{
		return true;
	}

	for (iter = begin(); iter != end(); ++iter)
	{
		if (iter->second.Includes(id))
		{
			return true;
		}
	}

	return false;
}

/**
 * @brief Get whether a body is a descendent of this body
 *
 * Get whether the given body is a descendent of this body
 *
 * @returns true if other is a descendent of this body
 *
 **/
bool
Body::Includes(Body const &other) const
{
	return Includes(other.Id());
}

/**
 * @brief Get body translation
 *
 * Get the body translation
 *
 * @returns the body translation
 *
 **/
Eigen::Vector3d
Body::Translation() const
{
	return Eigen::Vector3d(dBodyGetPosition(d_id));
}

/**
 * @brief Get body local translation
 *
 * Get the body local translation (with respect to its parent).
 *
 * @note This value is cached and is only calculated when the wrapper is
 * constructed!
 *
 * @returns the body local translation
 *
 **/
Eigen::Vector3d const &
Body::LocalTranslation() const
{
	return d_localTranslation;
}

/**
 * @brief Get body local rotation
 *
 * Get the body local rotation (with respect to its parent).
 *
 * @note This value is cached and is only calculated when the wrapper is
 * constructed!
 *
 * @returns the body local rotation
 *
 **/
Eigen::Matrix3d const &
Body::LocalRotation() const
{
	return d_localRotation;
}

/**
 * @brief Get body angular velocity
 *
 * Get the body angular velocity
 *
 * @returns the body angular velocity
 *
 **/
Eigen::Vector3d
Body::AngularVelocity() const
{
	return Eigen::Vector3d(dBodyGetAngularVel(d_id));
}

/**
 * @brief Get body linear velocity
 *
 * Get the body linear velocity
 *
 * @returns the body linear velocity
 *
 **/
Eigen::Vector3d
Body::LinearVelocity() const
{
	return Eigen::Vector3d(dBodyGetLinearVel(d_id));
}

/**
 * @brief Get body rotation quaternion
 *
 * Get the body rotation quaternion
 *
 * @returns the body rotation quaternion
 *
 **/
Eigen::Quaternion<double>
Body::Quaternion() const
{
	dReal const *quat = dBodyGetQuaternion(d_id);

	return Eigen::Quaternion<double>(quat[0], quat[1], quat[2], quat[3]);
}

/**
 * @brief Get body rotation matrix
 *
 * Get the body rotation matrix
 *
 * @returns the body rotation matrix
 *
 **/
Eigen::Matrix3d
Body::Rotation() const
{
	Eigen::Matrix3d ret;
	dReal const *rot = dBodyGetRotation(d_id);

	ret << rot[0], rot[4], rot[8],
	       rot[1], rot[5], rot[9],
	       rot[2], rot[6], rot[10];

	return ret;
}

/**
 * @brief Set body translation
 * @param pos the translation
 *
 * Set the body translation
 *
 **/
void
Body::SetTranslation(Eigen::Vector3d const &pos)
{
	dBodySetPosition(d_id, pos.x(), pos.y(), pos.z());
}

/**
 * @brief Set body angular velocity
 * @param angvel the angular velocity
 *
 * Set the body angular velocity
 *
 **/
void
Body::SetAngularVelocity(Eigen::Vector3d const &angvel)
{
	dBodySetAngularVel(d_id, angvel.x(), angvel.y(), angvel.z());
}

/**
 * @brief Set body linear velocity
 * @param linvel the linear velocity
 *
 * Set the body linear velocity
 *
 **/
void
Body::SetLinearVelocity(Eigen::Vector3d const &linvel)
{
	dBodySetLinearVel(d_id, linvel.x(), linvel.y(), linvel.z());
}

/**
 * @brief Set body rotation quaternion
 * @param quat the rotation quaternion
 *
 * Set the body rotation quaternion
 *
 **/
void
Body::SetQuaternion(Eigen::Quaternion<double> const &quat)
{
	dQuaternion q = {quat.x(), quat.y(), quat.z(), quat.w()};

	dBodySetQuaternion(d_id, q);
}

/**
 * @brief Set body rotation matrix
 * @param rotation the rotation matrix
 *
 * Set the body rotation matrix
 *
 **/
void
Body::SetRotation(Eigen::Matrix3d const &rotation)
{
	dReal vals[] = {
		rotation(0), rotation(3), rotation(6), 0,
		rotation(1), rotation(4), rotation(7), 0,
		rotation(2), rotation(5), rotation(8), 0
	};

	dBodySetRotation(d_id, vals);
}

/**
 * @brief Get all descendents of the body
 * @param includeSelf include body in results
 *
 * Get all descendents of this body
 *
 * @returns all descendents
 *
 **/
vector<Body>
Body::Descendents(bool includeSelf)
{
	vector<Body> ret;

	Descendents(ret, includeSelf);
	return ret;
}

void
Body::Descendents(vector<Body> &bodies, bool includeSelf)
{
	if (includeSelf)
	{
		bodies.push_back(*this);
	}

	for (iterator iter = begin(); iter != end(); ++iter)
	{
		iter->second.Descendents(bodies, true);
	}
}

/**
 * @brief Get body geoms
 *
 * Get all the geoms of the body
 *
 * @returns all the geoms
 *
 **/
vector<dGeomID>
Body::Geoms() const
{
	vector<dGeomID> geoms;

	if (!d_id)
	{
		return geoms;
	}

	dGeomID geom = dBodyGetFirstGeom(d_id);

	while (geom)
	{
		geoms.push_back(geom);
		geom = dBodyGetNextGeom(geom);
	}

	return geoms;
}

/**
 * @brief Get begin iterator
 *
 * Get the begin iterator of the children of the body
 *
 * @returns the begin iterator
 *
 **/
Body::BaseType::iterator
Body::begin()
{
	FindChildren();
	return Body::BaseType::begin();
}

/**
 * @brief Get end iterator
 *
 * Get the end iterator of the children of the body
 *
 * @returns the end iterator
 *
 **/
Body::BaseType::iterator
Body::end()
{
	FindChildren();
	return Body::BaseType::end();
}

/**
 * @brief Get begin iterator
 *
 * Get the begin iterator of the children of the body
 *
 * @returns the begin iterator
 *
 **/
Body::BaseType::const_iterator
Body::begin() const
{
	const_cast<Body &>(*this).FindChildren();

	return Body::BaseType::begin();
}

/**
 * @brief Get end iterator
 *
 * Get the end iterator of the children of the body
 *
 * @returns the end iterator
 *
 **/
Body::BaseType::const_iterator
Body::end() const
{
	const_cast<Body &>(*this).FindChildren();

	return Body::BaseType::end();
}
