#include "module.hh"
#include <libwebots/plugin/modules/debug/debug.hh>
#include <libwebots/plugin/plugin.hh>

using namespace libwebots::plugin;

/**
 * @class libwebots::plugin::Module
 * @brief Base plugin module
 *
 * This class is the base module class. Modules are implementations of particlar
 * actions in a physics plugin. This can range from a simple module to move
 * ODE objects to modules that implement particular physics
 * (such as hydrodynamics). Modules that need to hook into particular callbacks
 * provided by webots (such as PhysicsModule::Step() or
 * PhysicsModule::Collide()) should inherit from the PhysicsModule class instead
 * of from this class.
 *
 * Modules can have dependencies on other modules. For example, each module is
 * automatically dependent on the libwebots::plugin::modules::Debug module
 * which provides convenient
 * debugging facilities to other modules. Dependencies are implemented using
 * an interface which provides access to specific features of a module. For
 * example, the libwebots::plugin::modules::Debug module exposes the
 * libwebots::plugin::modules::Interface<Debug> class which allows
 * sending debugging messages.
 *
 * To register a module dependency, a module can implement a static method,
 * RegisterDependencies(), which is called once when registering a module type
 * on the Plugin. In this method, you can call Plugin::Register<ModuleName>() to
 * register module types on which you depend. Then, you can use
 * GetInterface<ModuleName>() once your module is constructed to retrieve the
 * modules::interfaces::Interface class providing an interface to that particular module.
 *
 * If this is all very confusing, then have a look at the tutorial which
 * provides a more detailed description of the libwebots plugin framework.
 *
 **/

/**
 * @brief Create module object
 * @param registeredDependencies map of registered dependencies
 *
 * Create new module object
 *
 **/
Module::Module(MapByTypeID const &registeredDependencies)
:
	d_registeredDependencies(registeredDependencies)
{
}

Module::~Module()
{
}

/**
 * @brief Register dependencies
 *
 * This static method can be overriden by particular modules to register
 * modules on which this module depends.
 *
 * @par Example
 * @code
 * void MyModule::RegisterDependencies()
 * {
 *   Plugin::Instance().Register<MyDependency>();
 * }
 * @endcode
 *
 **/
void
Module::RegisterDependencies()
{
	Plugin::Instance().Register<modules::Debug>();
}

