#include "constraint.hh"
#include <libwebots/communication/server.hh>
#include <libwebots/communication/messages.hh>
#include <algorithm>
#include <iostream>

#include <libwebots/plugin/utils.hh>
#include <libwebots/plugin/plugin.hh>

#include <libwebots/config.hh>

using namespace libwebots::plugin::modules;
using namespace std;

Constraint::Constraint(dWorldID world,
                       dSpaceID space,
                       dJointGroupID jointGroup,
                       MapByTypeID const &dependencies)
:
	PhysicsModule(world, space, jointGroup, dependencies),
	interfaces::Debugable(GetInterface<Debug>(),"Constraint"),
	d_previousTime(0)
{
	libwebots::communication::Server &server = libwebots::communication::Server::Instance();

	server.AddHandler(*this, messages::constraint::constraint, &Constraint::Handler);

	Info() << "Initialized module." << endl;
}

Constraint::~Constraint()
{
	for (map<string, Monitor *>::iterator iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		if (dJointGetFeedback(iter->second->Joint) == &(iter->second->JointFeedback))
		{
			dJointSetFeedback(iter->second->Joint, 0);
		}

		delete iter->second;
	}

	d_monitors.clear();

	for (vector<Item *>::iterator iter = d_constraints.begin(); iter != d_constraints.end(); ++iter)
	{
		Destroy(*iter, false);
	}
}

void
Constraint::Destroy(Item *item, bool remove)
{
	vector<Item *>::iterator iter;

	if (remove)
	{
		iter = find(d_constraints.begin(), d_constraints.end(), item);
	}

	if (!remove || iter != d_constraints.end())
	{
		vector<Item *>::iterator timeout;
		timeout = find(d_timeouts.begin(), d_timeouts.end(), item);

		if (timeout != d_timeouts.end())
		{
			d_timeouts.erase(timeout);
		}

		map<string, Monitor *>::iterator monitor;

		monitor = d_monitors.find(item->Name);

		if (monitor != d_monitors.end())
		{
			if (dJointGetFeedback(item->Joint) == &(monitor->second->JointFeedback))
			{
				dJointSetFeedback(item->Joint, 0);
			}

			delete monitor->second;

			d_monitors.erase(monitor);
		}

		if (item->DestroyJoint)
		{
			dJointDestroy(item->Joint);
		}

		if (remove)
		{
			d_constraints.erase(iter);
		}

		delete item;
	}
}

void
#ifdef ENABLE_JOINT_ENABLE
Constraint::Enable(Item *item, bool enable)
#else
Constraint::Enable(Item *, bool )
#endif
{
#ifdef ENABLE_JOINT_ENABLE
	std::vector<dJointID>::iterator monitor;
	monitor = find(d_monitors.begin(), d_monitors.end(), item->Joint);

	if (enable)
	{
		dJointEnable(item->Joint);
	}
	else
	{
		dJointDisable(item->Joint);
	}
#else
	Error() << "dJointEnable/dJointDisable is not available in this version of webots/ODE" << endl;
#endif
}

void
Constraint::Step()
{
	double time = dWebotsGetTime();

	vector<Item *>::iterator iter = d_timeouts.begin();

	while (iter != d_timeouts.end())
	{
		if (time > (*iter)->Timeout)
		{
			if ((*iter)->DestroyOnTimeout)
			{
				Destroy(*iter, true);
			}
			else
			{
				Enable(*iter, false);
			}

			iter = d_timeouts.erase(iter);
		}
		else
		{
			++iter;
		}
	}
}

void
Constraint::StepEnd()
{
	if (d_monitors.size() == 0)
	{
		return;
	}

	map<string, Monitor *>::iterator iter;
	communication::Server &server = communication::Server::Instance();

	double t = dWebotsGetTime() / 1000.0;
	double step = t - d_previousTime;

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		Monitor *monitor = iter->second;

		messages::constraint::JointMonitor monmsg;
		monmsg.set_name(monitor->Name);

		int type = dJointGetType(monitor->Joint);

		double position = 0;
		double velocity = 0;
		dVector3 axis = {0, 0, 0};

		switch (type)
		{
			case dJointTypeHinge:
				position = dJointGetHingeAngle(monitor->Joint);
				velocity = dJointGetHingeAngleRate(monitor->Joint);

				dJointGetHingeAxis(monitor->Joint, axis);
			break;
			case dJointTypeSlider:
				position = dJointGetSliderPosition(monitor->Joint);
				velocity = dJointGetSliderPositionRate(monitor->Joint);

				dJointGetSliderAxis(monitor->Joint, axis);
			break;
		}

		monmsg.set_position(position);
		monmsg.set_velocity(velocity);
		monmsg.set_acceleration((velocity - monitor->PreviousVelocity) / step);

		monitor->PreviousVelocity = velocity;

		for (size_t i = 0; i < 3; ++i)
		{
			monmsg.add_force1(monitor->JointFeedback.f1[i]);
			monmsg.add_torque1(monitor->JointFeedback.t1[i]);
			monmsg.add_force2(monitor->JointFeedback.f2[i]);
			monmsg.add_torque2(monitor->JointFeedback.t2[i]);

			monmsg.add_axis(axis[i]);
		}

		server.Send(communication::Messages::Create(messages::constraint::jointmonitor, monmsg, iter->first));
	}

	d_previousTime = t;
}

bool
Constraint::FindBody(string const &name, dBodyID &body)
{
	body = dWebotsGetBodyFromDEF(name.c_str());

	if (!body)
	{
		Error() << "Could not find body `" << name << "'" << endl;
		return false;
	}

	return true;
}

void
Constraint::AddMonitor(string const &name)
{
	dBodyID body;

	if (!FindBody(name, body))
	{
		return;
	}

	dJointID joint = Utils::FindWebotsJoint(body);

	if (!joint)
	{
		Error() << "could not find joint for body "
		          << "`" << name << "'"
		          << endl;
		return;
	}

	AddMonitor(name, joint);
}

void
Constraint::AddMonitor(string const &name,
                       dJointID      joint)
{
	if (d_monitors.find(name) != d_monitors.end())
	{
		return;
	}

	Monitor *monitor = new Monitor();

	monitor->Name = name;
	monitor->Joint = joint;
	monitor->PreviousVelocity = 0;

	int type = dJointGetType(joint);

	switch (type)
	{
		case dJointTypeHinge:
			monitor->PreviousVelocity = dJointGetHingeAngleRate(joint);
		break;
		case dJointTypeSlider:
			monitor->PreviousVelocity = dJointGetSliderPositionRate(joint);
		break;
	}

	if (!dJointGetFeedback(joint))
	{
		dJointSetFeedback(joint, &monitor->JointFeedback);
	}

	d_monitors[name] = monitor;
}

void
Constraint::Add(string const &name,
                dBodyID       body,
                dJointID      joint,
                double        timeout,
                bool          destroyOnTimeout,
                bool          destroyJoint,
                bool          monitor)
{
	Item *item = new Item();

	item->Name = name;
	item->Body = body;
	item->Joint = joint;
	item->Timeout = timeout;
	item->DestroyOnTimeout = destroyOnTimeout;
	item->DestroyJoint = destroyJoint;

	if (monitor)
	{
		AddMonitor(name, joint);
	}

	if (timeout >= 0)
	{
		item->Timeout += dWebotsGetTime();
		d_timeouts.push_back(item);
	}

	d_constraints.push_back(item);
}

void
Constraint::Handler(messages::constraint::FixInWorld const &message)
{
	dBodyID body;

	if (!FindBody(message.body(), body))
	{
		return;
	}

	// Create new fixed joint
	dJointID joint = dJointCreateFixed(World(), 0);

	// Attach body and the static world to the joint
	dJointAttach(joint, body, 0);
	dJointSetFixed(joint);

	Add(message.body(),
	    body,
	    joint,
	    message.disable_timeout(),
	    message.destroy_on_timeout(),
	    true,
	    message.monitor());

}

void
Constraint::Handler(messages::constraint::FixInPlane const &message)
{
	dBodyID body;

	if (!FindBody(message.body(), body))
	{
		Error() << "Could not find body '" << message.body() << "'."
		        << std::endl;
		return;
	}
	Info() << "Adding a new plane constraint for body '" << message.body()
	       << "'." << endl;

	dJointID joint = dJointCreateFixed(World(), 0);

	dBodyID con1 = dBodyCreate(World());
	dBodyID con2 = dBodyCreate(World());
	dBodyID con3 = 0;

	dReal const *pos = dBodyGetPosition(body);

	dBodySetPosition(con1, pos[0], pos[1], pos[2]);
	dBodySetPosition(con2, pos[0], pos[1], pos[3]);

	dJointID slider1 = dJointCreateSlider(World(), 0);
	dJointID slider2 = dJointCreateSlider(World(), 0);
	dJointID hinge = 0;

	// Attach the dummy to the fixed world
	dJointAttach(joint, con1, 0);
	dJointSetFixed(joint);

	dJointAttach(slider1, con1, con2);

	dVector3 axis1 = {0, 1, 0};
	dVector3 axis2 = {0, 0, 1};

	if (message.axis1_size() == 3)
	{
		for (size_t i = 0; i < 3; ++i)
		{
			axis1[i] = message.axis1(i);
		}
	}

	if (message.axis2_size() == 3)
	{
		for (size_t i = 0; i < 3; ++i)
		{
			axis2[i] = message.axis2(i);
		}
	}
	dNormalize3(axis1);
	dNormalize3(axis2);

	if (message.has_canrotate() && message.canrotate())
	{
		Info() << "--- Allowing to rotate in the plane." << endl;
		con3 = dBodyCreate(World());

		dBodySetPosition(con3, pos[0], pos[1], pos[2]);

		hinge = dJointCreateHinge(World(), 0);

		dJointAttach(slider2, con2, con3);
		dJointAttach(hinge, con3, body);

		dJointSetHingeAnchor(hinge, pos[0], pos[1], pos[2]);

		dVector3 hingeAxis;
		dCROSS(hingeAxis, =, axis1, axis2);
		dNormalize3(hingeAxis);

		dJointSetHingeAxis(hinge, hingeAxis[0], hingeAxis[1], hingeAxis[2]);

		if (message.rotation_stops_size() == 2)
		{
			Info() << "--- Adding stop position [ "
			       << message.rotation_stops(0) << " , "
			       << message.rotation_stops(1) << " ] for this rotation."
			       << endl;
			dJointSetHingeParam (hinge, dParamLoStop, message.rotation_stops(0));
			dJointSetHingeParam (hinge, dParamHiStop, message.rotation_stops(1));
		}
		else
		{
			dJointSetHingeParam (hinge, dParamLoStop, -1.5);
			dJointSetHingeParam (hinge, dParamHiStop,  1.5);
		}
	}
	else
	{
		dJointAttach(slider2, con2, body);
	}

	dJointSetSliderAxis(slider1, axis1[0], axis1[1], axis1[2]);
	dJointSetSliderAxis(slider2, axis2[0], axis2[1], axis2[2]);

	Add(message.body() + "::constraint1",
	    body,
	    slider1,
	    message.disable1_timeout(),
	    message.destroy1_on_timeout(),
	    true,
	    message.monitor1());

	Add(message.body() + "::constraint2",
	    body,
	    slider2,
	    message.disable2_timeout(),
	    message.destroy2_on_timeout(),
	    true,
	    message.monitor2());

	if (hinge)
	{
		Add(message.body() + "::rotationconstraint",
		    body,
		    hinge,
		    message.disablerotation_timeout(),
		    message.destroyrotation_on_timeout(),
		    true,
		    message.monitorrotation());
	}
}

void
Constraint::Handler(messages::constraint::FixInAxis const &message)
{
	dBodyID body;

	if (!FindBody(message.body(), body))
	{
		return;
	}

	dJointID joint = dJointCreateFixed(World(), 0);

	dReal const *pos = dBodyGetPosition(body);

	dBodyID con1 = dBodyCreate(World());
	dBodySetPosition(con1, pos[0], pos[1], pos[2]);

	// Attach the dummy to the fixed world
	dJointAttach(joint, con1, 0);
	dJointSetFixed(joint);

	dJointID slider1 = dJointCreateSlider(World(), 0);

	if (message.axis_size() == 3)
	{
		dJointSetSliderAxis(slider1, message.axis(0), message.axis(1), message.axis(2));
	}
	else
	{
		dJointSetSliderAxis(slider1, 0, 1, 0);
	}

	dJointAttach(slider1, con1, body);

	Add(message.body() + "::constraint1",
	    body,
	    slider1,
	    message.disable_timeout(),
	    message.destroy_on_timeout(),
	    true,
	    message.monitor());
}

void
Constraint::Handler(messages::constraint::Destroy const &message)
{
	dBodyID body;

	if (!FindBody(message.body(), body))
	{
		return;
	}

	vector<Item *>::iterator iter = d_constraints.begin();

	while (iter != d_constraints.end())
	{
		if ((*iter)->Body == body)
		{
			Destroy(*iter, false);

			iter = d_constraints.erase(iter);
		}
		else
		{
			++iter;
		}
	}
}

void
Constraint::Handler(messages::constraint::Enable const &message)
{
	dBodyID body;

	if (!FindBody(message.body(), body))
	{
		return;
	}

	vector<Item *>::iterator iter;

	for (iter = d_constraints.begin(); iter != d_constraints.end(); ++iter)
	{
		if ((*iter)->Body == body)
		{
			Enable(*iter, message.enable());
		}
	}
}

void
Constraint::Handler(messages::constraint::MonitorJoint const &message)
{
	AddMonitor(message.body());
}

void
Constraint::Handler(messages::constraint::SetVelocity const &message)
{
	dBodyID body;

	if (!FindBody(message.body(), body))
	{
		return;
	}

	dJointID joint = Utils::FindWebotsJoint(body);

	if (!joint)
	{
		Error() << "Could not find compatible joint for "
		        << "`" << message.body() << "'"
		        << endl;
		return;
	}

	int type = dJointGetType(joint);

	if (type == dJointTypeHinge)
	{
		dJointSetHingeParam(joint, dParamVel, message.velocity());
	}
	else if (type == dJointTypeSlider)
	{
		dJointSetSliderParam(joint, dParamVel, message.velocity());
	}
}

void
Constraint::Handler(messages::constraint::Constraint const &message)
{
	switch (message.action())
	{
		case messages::constraint::Action::FixInWorld:
			if (message.has_fixinworld())
			{
				Handler(message.fixinworld());
			}
			else
			{
				Error() << "Missing 'fixinworld' field" << endl;
			}
		break;
		case messages::constraint::Action::FixInPlane:
			if (message.has_fixinplane())
			{
				Handler(message.fixinplane());
			}
			else
			{
				Error() << "Missing 'fixinplane' field" << endl;
			}
		break;
		case messages::constraint::Action::MonitorJoint:
			if (message.has_monitorjoint())
			{
				Handler(message.monitorjoint());
			}
			else
			{
				Error() << "Missing 'monitorjoint' field" << endl;
			}
		break;
		case messages::constraint::Action::Destroy:
			if (message.has_destroy())
			{
				Handler(message.destroy());
			}
			else
			{
				Error() << "Missing 'destroy' field" << endl;
			}
		break;
		case messages::constraint::Action::Enable:
			if (message.has_enable())
			{
				Handler(message.enable());
			}
			else
			{
				Error() << "Missing 'enable' field" << endl;
			}
		break;
		case messages::constraint::Action::SetVelocity:
			if (message.has_setvelocity())
			{
				Handler(message.setvelocity());
			}
			else
			{
				Error() << "Missing 'setvelocity' field" << endl;
			}
		break;
		case messages::constraint::Action::FixInAxis:
			if (message.has_fixinaxis())
			{
				Handler(message.fixinaxis());
			}
			else
			{
				Error() << "Missing 'fixinaxis' field" << endl;
			}
		break;
		default:
		break;
	}
}
