#include "contact.hh"
#include <libwebots/communication/server.hh>
#include <libwebots/communication/messages.hh>
#include <algorithm>
#include <iostream>

#include <libwebots/plugin/utils.hh>
#include <libwebots/plugin/plugin.hh>

#include <libwebots/config.hh>

using namespace libwebots::plugin::modules;
using namespace std;

Contact::Contact(dWorldID world,
                 dSpaceID space,
                 dJointGroupID jointGroup,
                 Module::MapByTypeID const &dependencies)
:
	PhysicsModule(world, space, jointGroup, dependencies),
	Debugable(GetInterface<Debug>(),"Contact")
{
	libwebots::communication::Server &server = libwebots::communication::Server::Instance();

	server.AddHandler(*this, messages::contact::contact, &Contact::Handler);

	Info() << "module initialized." << std::endl;
}

Contact::~Contact()
{
	for (vector<Monitor *>::iterator iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		delete *iter;
	}
}

bool
Contact::SpaceMatch(dGeomID space, dGeomID g1)
{
	return dGeomIsSpace(space) && dSpaceQuery((dSpaceID)space, g1);
}

Contact::Monitor *
Contact::MatchMonitor(dGeomID g1, dGeomID g2)
{
	vector<Monitor *> ret;
	vector<Monitor *>::iterator iter;

	// Handle spaces
	for (iter = d_spaces.begin(); iter != d_spaces.end(); ++iter)
	{
		Monitor *monitor = *iter;

		if ((SpaceMatch(monitor->Geom1, g1) && (!monitor->Geom2 || g2 == monitor->Geom2)) ||
		    (monitor->Geom2 && SpaceMatch(monitor->Geom2, g1) && g2 == monitor->Geom1))
		{
			return monitor;
		}
		else if ((SpaceMatch(monitor->Geom1, g2) && (!monitor->Geom2 || g1 == monitor->Geom2)) ||
		         (monitor->Geom2 && SpaceMatch(monitor->Geom2, g2) && g1 == monitor->Geom1))
		{
			return monitor;
		}
	}

	// Handle normal geoms
	pair<multimap<dGeomID, Monitor *>::iterator, multimap<dGeomID, Monitor *>::iterator> found;

	found = d_monitorMap.equal_range(g1);

	while (found.first != found.second)
	{
		Monitor *monitor;

		monitor = found.first->second;

		if (!monitor->Geom2 || monitor->Geom2 == g2)
		{
			return monitor;
		}

		++found.first;
	}

	found = d_monitorMap.equal_range(g2);

	while (found.first != found.second)
	{
		Monitor *monitor;

		monitor = found.first->second;

		if (!monitor->Geom2 || monitor->Geom2 == g1)
		{
			return monitor;
		}

		++found.first;
	}

	return 0;
}

bool
Contact::Collide(dGeomID g1, dGeomID g2)
{
	Monitor *monitor;

	monitor = MatchMonitor(g1, g2);

	if (!monitor)
	{
		return false;
	}

	size_t n = dCollide(g1, g2, MaxContacts::Value, &(monitor->Contacts[0].geom), sizeof(dContact));

	if (!monitor->CustomModel && monitor->Monitor)
	{
		// Just send only the positions
		for (size_t i = 0; i < n; ++i)
		{
			modules::messages::contact::Monitor msg;
			msg.set_name(monitor->Name1);

			if (monitor->Geom2)
			{
				msg.set_other(monitor->Name2);
			}

			msg.add_position(monitor->Contacts[i].geom.pos[0]);
			msg.add_position(monitor->Contacts[i].geom.pos[1]);
			msg.add_position(monitor->Contacts[i].geom.pos[2]);
		}

		return false;
	}
	else if (monitor->CustomModel)
	{
		// Create contact joints etc.
		monitor->NumContacts = n;

		for (size_t i = 0; i < n; ++i)
		{
			if(monitor->Mode & dContactFDir1){
				ComputeDirection(monitor,i);
			}
			SetupSurface(monitor, i);

			dJointID joint = dJointCreateContact(World(),
			                                     JointGroup(),
			                                     &(monitor->Contacts[i]));

			dJointAttach(joint, dGeomGetBody(g1), dGeomGetBody(g2));
			//resetting the feedback structure to avoid NaN.
			memset(&(monitor->JointFeedback[i]),0,sizeof(dJointFeedback));
			dJointSetFeedback(joint, &(monitor->JointFeedback[i]));

		}

		return true;
	}

	return false;
}

void
Contact::ComputeDirection(Monitor * monitor, size_t i){
	dVector3 & fdir = monitor->Contacts[i].fdir1;
	dVector3 & normal = monitor->Contacts[i].geom.normal;
	dGeomID & g = monitor->Contacts[i].geom.g1;

	if(dGeomIsSpace(g)){
		monitor->Mode &= ~dContactFDir1;
		return;
	}

	const dReal * ori = dGeomGetRotation(g);
	dVector3 unit,dirBody;
	unit[0] = 1;
	unit[1] = 0;
	unit[2] = 0;
	double normDir(0.0);
	while(normDir < 1e-5){
		dMULTIPLY0_331(dirBody,ori,unit);
		dCROSS(fdir,=,normal,dirBody);
		normDir = dLENGTH(fdir);
		if(unit[2] == 1){
			break;
		} else {
			unit[2] = 1;
			unit[0] = 0;
		}
	}

	fdir[0] /= normDir;
	fdir[1] /= normDir;
	fdir[2] /= normDir;
}

void
Contact::SetupSurface(Monitor *monitor, size_t i)
{
	dSurfaceParameters *p;

	p = &(monitor->Contacts[i].surface);

	p->mode = monitor->Mode;
	p->mu = monitor->CoulombFriction;
	p->bounce = monitor->Bounce;
	p->bounce_vel = monitor->BounceVelocity;
	p->slip1 = monitor->ForceDependentSlip;
	p->slip2 = monitor->ForceDependentSlip;
	p->soft_cfm = monitor->SoftCFM;
	p->soft_erp = monitor->SoftERP;
}

void
Contact::StepEnd()
{
	vector<Monitor *>::iterator iter;

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		Monitor *monitor;

		monitor = *iter;

		if (monitor->Monitor)
		{
			modules::messages::contact::Monitor msg;
			msg.set_name(monitor->Name1);
			if (monitor->Geom2)
			{
				msg.set_other(monitor->Name2);
			}

			for (size_t i = 0; i < monitor->NumContacts; ++i)
			{
				msg.add_position(monitor->Contacts[i].geom.pos[0]);
				msg.add_position(monitor->Contacts[i].geom.pos[1]);
				msg.add_position(monitor->Contacts[i].geom.pos[2]);

				msg.add_force(monitor->JointFeedback[i].f1[0]);
				msg.add_force(monitor->JointFeedback[i].f1[1]);
				msg.add_force(monitor->JointFeedback[i].f1[2]);

				msg.add_torque(monitor->JointFeedback[i].t1[0]);
				msg.add_torque(monitor->JointFeedback[i].t1[1]);
				msg.add_torque(monitor->JointFeedback[i].t1[2]);

			}

			communication::Server::Instance().Send(communication::Messages::Create(messages::contact::contactmonitor,msg));
			if(monitor->NumContacts)
			{
				Info() << "Found " << monitor->NumContacts << " between "
				       << monitor->Name1 << " and " << monitor->Name2
				       << std::endl;
			}
		}

		if (!monitor->DrawContact)
		{
			monitor->NumContacts = 0;
		}
	}
}

bool
Contact::FindGeom(string const &name, dGeomID &geom)
{
	geom = dWebotsGetGeomFromDEF(name.c_str());

	if (!geom)
	{
		dWebotsConsolePrintf("[Contact] Could not find body '%s'\n", name.c_str());
		return false;
	}

	return true;
}

void
Contact::Handler(messages::contact::Contact const &message)
{
	dGeomID g1;
	dGeomID g2 = 0;
	bool newmonitor = false;

	if (!FindGeom(message.body1(), g1))
	{
		Error() << "Could not find body " << message.body1() << std::endl;
		return;
	}

	if (message.has_body2() && !FindGeom(message.body2(), g2))
	{
		Error() << "Could not find body " << message.body2() << std::endl;
		return;
	}

	/* Match existing */
	vector<Monitor *>::iterator iter;
	Monitor *monitor = 0;

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		if ((*iter)->Geom1 == g1 && (*iter)->Geom2 == g2)
		{
			monitor = *iter;
			break;
		}
	}

	if (!monitor)
	{
		monitor = new Monitor();
		newmonitor = true;

		monitor->Geom1 = g1;
		monitor->Geom2 = g2;

		monitor->Mode = 0;
		monitor->CustomModel = false;
		monitor->Monitor = false;
		monitor->DrawContact = false;

		monitor->Name1 = message.body1();

		if (message.has_body2())
		{
			monitor->Name2 = message.body2();
		}
	}

	if (message.has_monitor())
	{
		monitor->Monitor = message.monitor();
	}

	if (message.has_draw())
	{
		monitor->DrawContact = message.draw();
	}

	if (message.has_bounce() && message.has_bouncevelocity())
	{
		monitor->Mode |= dContactBounce;

		monitor->Bounce = message.bounce();
		monitor->BounceVelocity = message.bouncevelocity();

		monitor->CustomModel = true;
	}

	if (message.has_coulombfriction())
	{
		monitor->Mode |= dContactApprox1;
		monitor->CoulombFriction = message.coulombfriction();

		monitor->CustomModel = true;
	}

	if (message.has_forcedependentslip() && message.forcedependentslip() > 0)
	{
		monitor->Mode |= dContactSlip1 | dContactSlip2;
		monitor->CustomModel = true;

		monitor->ForceDependentSlip = message.forcedependentslip();
	}

	if (message.has_softerp())
	{
		monitor->Mode |= dContactSoftERP;
		monitor->CustomModel = true;

		monitor->SoftERP = message.softerp();
	}

	if (message.has_softcfm())
	{
		monitor->Mode |= dContactSoftCFM;
		monitor->CustomModel = true;

		monitor->SoftCFM = message.softcfm();
	}

	if(message.has_computedirection())
	{
		monitor->Mode |= dContactFDir1;
	}

	if (newmonitor)
	{
		d_monitors.push_back(monitor);

		if (!dGeomIsSpace(g1))
		{
			d_monitorMap.insert(pair<dGeomID, Monitor *>(g1, monitor));
		}

		if (dGeomIsSpace(g1) || (g2 && dGeomIsSpace(g2)))
		{
			d_spaces.push_back(monitor);
		}

		if (g2 && !dGeomIsSpace(g2))
		{
			d_monitorMap.insert(pair<dGeomID, Monitor *>(g2, monitor));
		}
	}

}

void
Contact::DrawContact(dContact const &contact, dJointFeedback const &feedback, bool vec)
{
	double const *pos = contact.geom.pos;

	if (vec)
	{
		//double const *normal = contact.geom.normal;

		glVertex3f(pos[0], pos[1], pos[2]);
		glVertex3f(pos[0] + feedback.f1[0] / 1000,
		           pos[1] + feedback.f1[1] / 1000,
		           pos[2] + feedback.f1[2] / 1000);
	}
	else
	{
		glVertex3f(pos[0] - 0.01, pos[1], pos[2] - 0.01);
		glVertex3f(pos[0] - 0.01, pos[1], pos[2] + 0.01);
		glVertex3f(pos[0] + 0.01, pos[1], pos[2] + 0.01);
		glVertex3f(pos[0] + 0.01, pos[1], pos[2] - 0.01);
	}
}

void
Contact::Draw()
{
	vector<Monitor *>::iterator iter;

	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);

	glColor3f(0, 0, 1);

	glDisable(GL_DEPTH_TEST);

	glBegin(GL_QUADS);

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		Monitor *monitor;

		monitor = *iter;

		if (!monitor->DrawContact)
		{
			continue;
		}

		for (size_t i = 0; i < monitor->NumContacts; ++i)
		{
			DrawContact(monitor->Contacts[i], monitor->JointFeedback[i], false);
		}
	}

	glEnd();

	//glEnable(GL_DEPTH_TEST);
	glLineWidth(2.0);

	glBegin(GL_LINES);

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		Monitor *monitor;

		monitor = *iter;

		if (!monitor->DrawContact)
		{
			continue;
		}

		for (size_t i = 0; i < monitor->NumContacts; ++i)
		{
			DrawContact(monitor->Contacts[i], monitor->JointFeedback[i], true);
		}



	}

	glEnd();

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		Monitor *monitor;

		monitor = *iter;

		for (size_t i = 0; i < monitor->NumContacts; ++i)
		{
			if(monitor->Contacts[i].surface.mode & dContactFDir1 == 0)
			{
				continue;
			}

			double const *pos = monitor->Contacts[i].geom.pos;
			dVector3 & dir    = monitor->Contacts[i].fdir1;
			glBegin(GL_LINES);
			glVertex3f(pos[0],pos[1],pos[2]);
			glVertex3f(pos[0] + dir[0],
			           pos[1] + dir[1],
			           pos[2] + dir[2]);
			glEnd();
		}
	}


	glLineWidth(1.0);
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);

	for(iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		(*iter)->NumContacts = 0;
	}

}
