#include "debug.hh"

#include <libwebots/communication/server.hh>
#include <libwebots/communication/messages.hh>
#include <libwebots/plugin/plugin.hh>
#include <iostream>

using namespace std;
using namespace libwebots::plugin;
using namespace libwebots::plugin::modules;

modules::messages::debug::Type::Values Debug::s_defaultType = modules::messages::debug::Type::Error;

Debug::Debug(Module::MapByTypeID const &dependencies)
:
	Module(dependencies),
	d_enabled(s_defaultType),
	d_redirection(messages::debug::Redirection::Webots)
{
	libwebots::communication::Server &server = libwebots::communication::Server::Instance();

	server.AddHandler(*this, messages::debug::debug, &Debug::Handler);

	Send(messages::debug::Type::Info, "[Debug] Initialized module");
}

void
Debug::Handler(messages::debug::Debug const &message)
{
	switch (message.action())
	{
		case messages::debug::Action::Enable:
			if (!message.has_type())
			{
				Send(messages::debug::Type::Error, "[Debug] Missing type for Enable action");
			}
			else
			{
				d_enabled = static_cast<messages::debug::Type::Values>(d_enabled | message.type());
			}
		break;
		case messages::debug::Action::Disable:
			if (!message.has_type())
			{
				Send(messages::debug::Type::Error, "[Debug] Missing type for Disable action");
			}
			else
			{
				d_enabled = static_cast<messages::debug::Type::Values>(d_enabled & ~message.type());
			}
		break;
		case messages::debug::Action::Redirect:
			if (!message.has_redirection())
			{
				Send(messages::debug::Type::Error, "[Debug] Missing redirection for Redirect action");
			}
			else
			{
				d_redirection = message.redirection();
			}
		break;
	}
}

void
Debug::Send(messages::debug::Type::Values type, string const &message)
{
	if (!(d_enabled & type))
	{
		return;
	}

	string msg;

	if (message.empty() || message.at(message.size() - 1) != '\n')
	{
		msg = message + "\n";
	}
	else
	{
		msg = message;
	}

	switch (d_redirection)
	{
		case messages::debug::Redirection::Terminal:
			cerr << msg;
		break;
		case messages::debug::Redirection::Webots:
			dWebotsConsolePrintf("%s", msg.c_str());
		break;
		case messages::debug::Redirection::Message:
		{
			communication::Server &server = communication::Server::Instance();

			messages::debug::Message ms;
			ms.set_type(type);
			ms.set_message(msg);

			server.Send(communication::Messages::Create(messages::debug::debugmessage, ms));
		}
		break;
	}
}

void
Debug::RegisterDependencies()
{
	/* Prevent registration of debug depending on itself because the
	   default Module::RegisterDependencies does that... */
}

void
Debug::SetDefaultType(messages::debug::Type::Values type)
{
	s_defaultType = type;
}

interfaces::Interface<Debug>::Interface(Debug &module)
:
	d_debug(module),
	d_buffer(this)
{
}

interfaces::Interface<Debug>::Interface(interfaces::Interface<Debug> const &other)
:
	std::stringbuf(),
	d_debug(other.d_debug),
	d_type(other.d_type),
	d_buffer(this)
{
}

interfaces::Interface<Debug> &
interfaces::Interface<Debug>::operator=(interfaces::Interface<Debug> const &other)
{
	d_debug = other.d_debug;
	d_type = other.d_type;

	return *this;
}

int
interfaces::Interface<Debug>::sync()
{
	d_debug.Send(d_type, str());
	str("");

	return 0;
}

std::ostream &
interfaces::Interface<Debug>::operator()(messages::debug::Type::Values type)
{
	d_type = type;
	return d_buffer;
}

interfaces::Debugable::Debugable(const Interface<Debug> & debug,
                                 const std::string & prefix)
	: d_debug(debug)
	, d_prefix(prefix){
	if(!d_prefix.empty() && d_prefix.at(0) != '['){
		d_prefix.insert(d_prefix.begin(),'[');
		d_prefix.append("] : ");
	}
}


