/**
 * \file joints.cc
 *
 *  \date Nov 23, 2011
 *  \author Alexandre Tuleu
 */

#include "joints/joints.hh"

#include <iostream>

using namespace libwebots::plugin::modules::jointmanager;
using namespace libwebots::plugin::modules::messages::jointmanager;


Joint::JointFactory Joint::s_factory;
Joint::JointODEFactory Joint::s_odeFactory;
Joint::ParameterFactory Joint::s_parameters;

void Joint::RegisterFactories()
{
	if(!s_factory.empty() && !s_parameters.empty() && !s_odeFactory.empty()){
		std::cerr << "[JointManager] : Weird initialization error."
		          << std::endl;
		return;
	}

	s_factory[JointDefinition::HINGE] = new Hinge();
	s_odeFactory[dJointTypeHinge] = new Hinge();
	s_factory[JointDefinition::FIXED] = new Fixed();
	s_odeFactory[dJointTypeFixed] = new Fixed();
	s_factory[JointDefinition::SLIDER] = new Slider();
	s_odeFactory[dJointTypeSlider] = new Slider();
	s_factory[JointDefinition::BALL] = new Ball();
	s_odeFactory[dJointTypeBall]   = new Ball();


	s_parameters[JointParameters::dParamLoStop]        = dParamLoStop;
	s_parameters[JointParameters::dParamHiStop]        = dParamHiStop;
	s_parameters[JointParameters::dParamVel]           = dParamVel;
	s_parameters[JointParameters::dParamFMax]          = dParamFMax;
	s_parameters[JointParameters::dParamFudgeFactor]   = dParamFudgeFactor;
	s_parameters[JointParameters::dParamBounce]        = dParamBounce;
	s_parameters[JointParameters::dParamCFM]           = dParamCFM;
	s_parameters[JointParameters::dParamStopERP]       = dParamStopERP;
	s_parameters[JointParameters::dParamStopCFM]       = dParamStopCFM;
	s_parameters[JointParameters::dParamSuspensionERP] = dParamSuspensionERP;
	s_parameters[JointParameters::dParamSuspensionCFM] = dParamSuspensionCFM;

}


int Joint::ParameterFromMessage(messages::jointmanager::JointParameters::Values parameter)
{
	ParameterFactory::const_iterator fi = s_parameters.find(parameter);
	if(fi == s_parameters.end()){
		return 0x442;
	}
	return fi->second;
}



std::auto_ptr<Joint>
Joint::Create(const messages::jointmanager::JointDefinition::Type type)
{
	JointFactory::const_iterator fi = s_factory.find(type);

	if(fi == s_factory.end()){
		return std::auto_ptr<Joint>(NULL);
	}

	return fi->second->Clone();
}



std::auto_ptr<Joint> Joint::Create(const Body & parent, const Body & child)
{


	unsigned int numJoint = dBodyGetNumJoints(parent.Id());
	dJointID id = 0;
	for(unsigned int i = 0; i < numJoint; ++i)
	{
		dJointID j = dBodyGetJoint(parent.Id(),i);

		if(dJointGetBody(j,0) == parent.Id() && dJointGetBody(j,1) == child.Id())
		{
			id = j;
			break;
		}

		if(dJointGetBody(j,1) == parent.Id() && dJointGetBody(j,0) == child.Id())
		{
			id = j;
			break;
		}
	}

	if(id == 0)
	{
		return std::auto_ptr<Joint>(NULL);
	}


	JointODEFactory::const_iterator fij = s_odeFactory.find(dJointGetType(id));

	if(fij == s_odeFactory.end())
	{
		return std::auto_ptr<Joint>(NULL);
	}

	std::auto_ptr<Joint> res = fij->second->Clone();

	res->SetID(id);

	//here the joint was not created by us, so we should never destroy it.
	res->SetManaged(false);

	return res;

}


Joint::~Joint()
{
	if(d_id != 0 && d_managed)
	{
		dJointDestroy(d_id);
	}
}


void Joint::Step(double ts){
	ComputeCoordinateAndRate(ts);
}


void Joint::SetAnchor(const Eigen::Vector3d & worldPosition)
{
	SetAnchor(worldPosition.data());
}



void Joint::SetAxis(const Eigen::Vector3d & worldPosition)
{
	SetAxis(worldPosition.data());
}



Joint::operator bool() const
{
	return d_id != 0;
}



Joint::Joint() :
	d_id(0),
	d_managed(true)
{
}



void Joint::Create(dWorldID world, dJointGroupID group)
{
	if(d_id) { return; }
	d_id = CreatePrimpl(world,group);
}

void Joint::Attach(const Body & parent, const Body & child)
{
	Attach(parent.Id(),child.Id());
}

void Joint::Attach(dBodyID parent, dBodyID child)
{
	if(!*this) { return; }
	dJointAttach(d_id,parent,child);
}

const Eigen::Vector3d& Joint::GetCoordinateConst() const{
	return d_coordinate;
}

const Eigen::Vector3d& Joint::GetCoordinateRateConst() const{
	return d_rate;
}

Eigen::Vector3d& Joint::GetCoordinate(){
	return d_coordinate;
}

Eigen::Vector3d& Joint::GetCoordinateRate(){
	return d_rate;
}

void Joint::SetID(dJointID id)
{
	if(d_id)
	{
		return;
	}
	d_id = id;
}

void Joint::SetManaged(bool managed)
{
	d_managed = managed;
}



void Fixed::SetAnchor(const dVector3)
{
}



void Fixed::SetAxis(const dVector3)
{
}



std::auto_ptr<Joint> Fixed::Clone() const
{
	return std::auto_ptr<Joint>(new Fixed(*this));
}



dJointID Joint::Id() const
{
	return d_id;
}


dBodyID Joint::Parent() const{
	return dJointGetBody(d_id,0);
}

dBodyID Joint::Child() const{
	return dJointGetBody(d_id,1);
}





