/**
 * \file jointmanager.cc
 *
 *  \date Nov 18, 2011
 *  \author Alexandre Tuleu
 */

#include <libwebots/communication/server.hh>
#include <libwebots/communication/messages.hh>
#include <libwebots/plugin/body.hh>

#include "jointmanager.hh"
#include "joints/joints.hh"
#include "passivedynamics.hh"

#include <libwebots/plugin/plugin.hh>

using namespace libwebots::plugin::modules;

void JointManager::RegisterDependencies(){
	Module::RegisterDependencies();
	jointmanager::Joint::RegisterFactories();

}

JointManager::JointManager(dWorldID world,
                           dSpaceID space,
                           dJointGroupID group,
                           Module::MapByTypeID const & dependencies)
:
	PhysicsModule(world,space,group,dependencies),
	interfaces::Debugable(GetInterface<Debug>(),"JointCreator")
//	d_displayAxis(false)
{
	libwebots::communication::Server &server =
		libwebots::communication::Server::Instance();

	server.AddHandler(*this, messages::jointmanager::jointmanager, &JointManager::Handler);

	Info() << "Initialized module." << std::endl;
}

JointManager::~JointManager()
{
	//clear monitor and passive dynamics first
	d_monitors.clear();
	d_passiveDynamics.clear();
}


void JointManager::Handler(messages::jointmanager::JointManager const & message)
{
	switch(message.action()){
	case messages::jointmanager::Action::Create :
		CreateHandler(message.create());
		break;
	case messages::jointmanager::Action::Destroy :
		DestroyHandler(message.destroy());
		break;
	case messages::jointmanager::Action::AddManaged :
		ManageHandler(message.addmanaged());
		break;
	case messages::jointmanager::Action::SetJointParameter :
		SetParameterHandler(message.setjointparameter());
		break;
	case messages::jointmanager::Action::SetPassiveDynamic :
		SetPassiveDynamicsHandler(message.setpassivedynamics());
		break;
	case messages::jointmanager::Action::AddMonitor :
		AddMonitorHandler(message.addmonitor());
		break;
	case messages::jointmanager::Action::DeleteMonitor :
		DeleteMonitorHandler(message.deletemonitor());
		break;
	default :
		Error() << "Received an unhandled action '"
		        << messages::jointmanager::Action::Values_Name(message.action())
		        << "'." << std::endl;
		break;
	}
}

void JointManager::CreateHandler(messages::jointmanager::Create const & message)
{
	//first find if the joint is uniquely identified

	if(!message.has_jointidentifier()){
		Error() << "You must provide a unique identifier at "
		        << "creation." << std::endl;
		return;
	}

	JointByIdentifier::const_iterator fi = d_joints.find(message.jointidentifier());

	if( fi != d_joints.end()){
		Error() << "The identifier '"
		        << message.jointidentifier() << "' is not unique." << std::endl;
		return;
	}

	if(!message.has_definition()){
		Error() << "You must provide a definition." << std::endl;
		return;
	}

	Body parent(message.definition().parent());

	if(!parent){
		Error() << "Parent body '" << message.definition().parent()
		        << "' is invalid." <<std::endl;
		return;
	}

	Body child(message.definition().child());

	if(!child){
		Error() << "Parent child'" << message.definition().child()
		        << "' is invalid." <<std::endl;
		return;
	}

	dVector3 position,axis,worldPosition,worldAxis;

	if(message.definition().has_position()){
		position[0] = message.definition().position().x();
		position[1] = message.definition().position().y();
		position[2] = message.definition().position().z();
	} else {
		memset(position,0,sizeof(dVector3));
	}

	if(message.definition().has_axis()){
		axis[0] = message.definition().axis().x();
		axis[1] = message.definition().axis().y();
		axis[2] = message.definition().axis().z();
	} else {
		axis[0] = 1.0;
		axis[1] = 0.0;
		axis[2] = 0.0;
	}

	//express everything in world coordinate system.

	switch (message.definition().referenceframe()) {
		case messages::jointmanager::JointDefinition::PARENT:
			dBodyGetRelPointPos(parent.Id(),
			                    position[0],
			                    position[1],
			                    position[2],
			                    worldPosition);
			dBodyVectorToWorld(parent.Id(),
			                   axis[0],
			                   axis[1],
			                   axis[2],
			                   worldAxis);
			break;
		case messages::jointmanager::JointDefinition::CHILD :
			dBodyGetRelPointPos(child.Id(),
			                    position[0],
			                    position[1],
			                    position[2],
			                    worldPosition);
			dBodyVectorToWorld(child.Id(),
			                   axis[0],
			                   axis[1],
			                   axis[2],
			                   worldAxis);
			break;
		case messages::jointmanager::JointDefinition::WORLD :
			memcpy(worldPosition,position,sizeof(dVector3));
			memcpy(worldAxis,axis,sizeof(dVector3));
			break;
		default:
			break;
	}

	std::auto_ptr<jointmanager::Joint> jPtr = jointmanager::Joint::Create(message.definition().type());
	std::string typeJointName(message.definition().Type_Name(message.definition().type()));
	if(!jPtr.get()){
		Error() << "Unsupported joint type '"
				<< typeJointName
				<< "." << std::endl;
	}

	JointPtr j(jPtr.release());


	j->Create(World(),0);
	j->Attach(parent,child);
	j->SetAnchor(worldPosition);
	j->SetAxis(worldAxis);

	d_joints[message.jointidentifier()] = j;

	if(message.definition().has_drawsize()){
		double size = std::fabs(message.definition().drawsize());
		if(size > 0.0){
			d_drawSizes[j] = size;
		}
	}


	Info() << "Created a new " << typeJointName << " joint ' "
	       << message.jointidentifier() <<"' between bodies '"
	       << message.definition().parent() << "' and '"
	       << message.definition().child() << "' at world position "
	       << "[ " << worldPosition[0] << " " << worldPosition[1] << " "
	       << worldPosition[2] << " ] and world axis [ "
	       << worldAxis[0] << " " << worldAxis[1] << " " << worldAxis[2] << " ]."
	       << std::endl;

}
void
JointManager::DestroyHandler(messages::jointmanager::Destroy const & message)
{
	if(!message.has_jointidentifier())
	{
		Error() << "You must provide an unique identifier "
		        << "to destroy." << std::endl;
		return;
	}

	JointByIdentifier::iterator fi = d_joints.find(message.jointidentifier());

	if( fi == d_joints.end())
	{
		Error() << "The identifier '"
		        << message.jointidentifier() << "' was not found." << std::endl;
		return;
	}

	//we destroy any passive dynamics associated

	PassiveDynamicsByIdentifier::iterator pv = d_passiveDynamics.find(message.jointidentifier());

	if( pv != d_passiveDynamics.end())
	{
		Info() << "Removing passive dynamic of joint '" << message.jointidentifier()
		       << "'" << std::endl;
		d_passiveDynamics.erase(pv);
	}

	//we destroy any monitors

	MonitorByIdentifier::iterator m = d_monitors.find(message.jointidentifier());

	if(m != d_monitors.end()){
		Info() << "Removing monitoring for joint '"
		       << message.jointidentifier() << "'" << std::endl;
		d_monitors.erase(m);
	}

	SizeByJoint::iterator size = d_drawSizes.find(fi->second);

	if(size != d_drawSizes.end()){
		d_drawSizes.erase(size);
	}

	d_joints.erase(fi);
	Info() << "Destroyed joint '"
	       << message.jointidentifier() << "'."
	       << std::endl;

}

void JointManager::ManageHandler(const messages::jointmanager::AddManagedJoint & message)
{
	if(!message.has_jointidentifier())
	{
		Error() << "You must provide and unique identifier to add a managed "
		        << "joint." << std::endl;
		return;
	}

	JointByIdentifier::iterator fi = d_joints.find(message.jointidentifier());

	if(fi != d_joints.end())
	{
		Error() << "Identifier '" << message.jointidentifier() << "' is already"
		        << " taken." << std::endl;
		return;
	}

	Body parent(message.parent());

	if(!parent){
		Error() << "Parent body '" << message.parent()
		        << "' is invalid." << std::endl;
		return;
	}

	Body child(message.child());

	if(!child){
		Error() << "Parent child'" << message.child()
		        << "' is invalid." << std::endl;
		return;
	}

	std::auto_ptr<jointmanager::Joint> j =
		jointmanager::Joint::Create(parent,child);

	if(!j.get()){
		Error() << "Joint between bodies '" << message.parent() <<"' and '"
		        << message.child() << "' could not be managed."
		        << std::endl;
		return;
	}

	Info() << "Current joint position is " << j->GetCoordinateConst() << std::endl;

	d_joints[message.jointidentifier()] = JointPtr(j.release());

	Info() << "Added a managed joint '" << message.jointidentifier()
	       <<"' between '" << message.parent() <<"' and '"
	       << message.child() <<"'."
	       << std::endl;
}



void
JointManager::SetParameterHandler(const messages::jointmanager::SetJointParameter & message)
{
	if(!message.has_jointidentifier())
	{
		Error() << "You must procide a joint identifier." << std::endl;
		return;
	}

	JointByIdentifier::iterator j = d_joints.find(message.jointidentifier());

	if( j == d_joints.end())
	{
		Error() << "Could not found any joint '" << message.jointidentifier()
		        << "'." << std::endl;
		return;
	}

	for(ListOfParameterValue::const_iterator pv = message.parameters().begin();
	    pv != message.parameters().end();
	    ++pv)
	{
		uint32_t dimension = 0;
		if(pv->has_dimension()){
			dimension = pv->dimension();
		}

		j->second->SetParameter(jointmanager::Joint::ParameterFromMessage(pv->parameter()),
		                        pv->value(),
		                        dimension);
		Info() << "Set for joint '" << message.jointidentifier()
		       <<"' parameter '"
		       << messages::jointmanager::JointParameters::Values_Name(pv->parameter())
		       << "' on dimension "
		       << dimension
		       << " to " << pv->value()
		       << std::endl;
	}

}

void
JointManager::SetPassiveDynamicsHandler(messages::jointmanager::SetPassiveDynamics const & message)
{
	if(!message.has_jointidentifier())
	{
		Error() << "You must provide an unique identifier."
		        << std::endl;
		return;
	}

	PassiveDynamicsByIdentifier::iterator pv = d_passiveDynamics.find(message.jointidentifier());

	if( pv != d_passiveDynamics.end())
	{
		pv->second->SetFromMessage(message);
		Info() << "Changed the PassiveDynamic attribute of joint '"
		       << message.jointidentifier() << "' to  : "
		       << *(pv->second) << "." << std::endl;
		return;
	}

	JointByIdentifier::iterator j = d_joints.find(message.jointidentifier());

	if( j == d_joints.end() )
	{
		Error() << "Unable to found joint '"
		        << message.jointidentifier() << "'." << std::endl;
		return;
	}

	PassiveDynamicsPtr npv(new jointmanager::PassiveDynamics((j->second)));

	npv->SetFromMessage(message);

	d_passiveDynamics[message.jointidentifier()] = npv;

	Info() << "Set new passive dynamics for joint '"
	       << message.jointidentifier() << "' with : "
	       << *(npv) << "." << std::endl;

}

void JointManager::AddMonitorHandler(messages::jointmanager::AddMonitor const & m){
	using namespace messages::jointmanager;

	if(!m.has_jointidentifier()){
		Error() << "You must provide a joint identifier"
				<< std::endl;
		return;
	}

	JointByIdentifier::iterator j = d_joints.find(m.jointidentifier());

	if(j == d_joints.end()){
		Error() << "Could not find any joint identified by '" << m.jointidentifier()
				<< "'" << std::endl;
		return;
	}

	AddMonitor::Type t(AddMonitor::JointFeedback);
	if(m.has_type()){
		t = m.type();
	}
	JointDefinition::Frame f(JointDefinition::WORLD);

	if(m.has_frame()){
		f = m.frame();
	}

	MonitorPtr monitor;
	switch(t){
	case AddMonitor::JointFeedback :
		monitor = MonitorPtr(new jointmanager::JointMonitorByFeedback(*(j->second),f));
		break;
	case AddMonitor::BodyTotalForceDifference :
		monitor = MonitorPtr(new jointmanager::JointMonitorByDifference(*(j->second),f));
		break;
	default :
		Error() << "Received and unhandled monitor type '"
		        << AddMonitor::Type_Name(t) << "'" << std::endl;
		return;
	}

	d_monitors[j->first] = monitor;

	Info() << "Added joint monitoring for joint id '" << m.jointidentifier()
	       << "'" << std::endl;

}

void JointManager::DeleteMonitorHandler(messages::jointmanager::DeleteMonitor const & m){
	if(!m.has_jointidentifier()){
		Error() << "You must provide a joint indientifier."
				<< std::endl;
		return;
	}

	MonitorByIdentifier::iterator mon = d_monitors.find(m.jointidentifier());

	if(mon == d_monitors.end()){
		Error() << "Could not find any monitored joint identified by '"
		        << m.jointidentifier() << "'" << std::endl;
		return;
	}


	d_monitors.erase(mon);
}

void JointManager::Step()
{

	for(JointByIdentifier::const_iterator j = d_joints.begin();
	    j != d_joints.end();
	    ++j){
		j->second->Step(Plugin::TimeStep());
	}

	for(PassiveDynamicsByIdentifier::iterator pv = d_passiveDynamics.begin();
	    pv != d_passiveDynamics.end();
	    ++pv)
	{
		pv->second->Step();
	}
}

void JointManager::StepEnd()
{
	for(MonitorByIdentifier::iterator m = d_monitors.begin();
	    m != d_monitors.end();
	    ++m){
		modules::messages::jointmanager::Monitor message;
		message.set_jointidentifier(m->first);
		m->second->PrepareMessage(message);

		communication::Server::Instance().Send(communication::Messages::Create(messages::jointmanager::jointmanagermonitor,message));
	}
}

void JointManager::Draw(){
//	return;
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	for(SizeByJoint::iterator j = d_drawSizes.begin();
	    j != d_drawSizes.end();
	    ++j){
		j->first->Draw(j->second);
	}
	glPopAttrib();
}
