/**
 * \file jointmanagerbuilder.cc
 *
 *  \date Nov 18, 2011
 *  \author Alexandre Tuleu
 */


#include "jointmanagerbuilder.hh"

#include <stdexcept>

using namespace libwebots::plugin::modules::messages::builders;
using namespace libwebots::communication;


Message
JointManager::CreateHingeJoint(uint32_t identifier,
                               std::string const & parent,
                               std::string const & child,
                               jointmanager::JointDefinition::Frame frame,
                               const Eigen::Vector3d & position,
                               const Eigen::Vector3d & axis,
                               double drawSize)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::Create);

	jointmanager::Create *create = m.mutable_create();

	create->set_jointidentifier(identifier);

	jointmanager::JointDefinition * definition = create->mutable_definition();

	definition->set_parent(parent);
	definition->set_child(child);
	definition->set_referenceframe(frame);
	definition->mutable_position()->set_x(position.x());
	definition->mutable_position()->set_y(position.y());
	definition->mutable_position()->set_z(position.z());

	definition->mutable_axis()->set_x(axis.x());
	definition->mutable_axis()->set_y(axis.y());
	definition->mutable_axis()->set_z(axis.z());

	definition->set_drawsize(drawSize);

	definition->set_type(jointmanager::JointDefinition::HINGE);

	return Messages::Create(jointmanager::jointmanager,m);
}

Message
JointManager::CreateSliderJoint(uint32_t identifier,
                                std::string const & parent,
                                std::string const & child,
                                jointmanager::JointDefinition::Frame frame,
                                const Eigen::Vector3d & axis)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::Create);

	jointmanager::Create *create = m.mutable_create();

	create->set_jointidentifier(identifier);

	jointmanager::JointDefinition * definition = create->mutable_definition();

	definition->set_parent(parent);
	definition->set_child(child);
	definition->set_referenceframe(frame);

	definition->mutable_axis()->set_x(axis.x());
	definition->mutable_axis()->set_y(axis.y());
	definition->mutable_axis()->set_z(axis.z());

	definition->set_type(jointmanager::JointDefinition::SLIDER);

	return Messages::Create(jointmanager::jointmanager,m);
}

Message
JointManager::CreateFixedJoint(uint32_t identifier,
				               std::string const & parent,
				               std::string const & child)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::Create);

	jointmanager::Create *create = m.mutable_create();

	create->set_jointidentifier(identifier);

	jointmanager::JointDefinition * definition = create->mutable_definition();

	definition->set_parent(parent);
	definition->set_child(child);

	definition->set_type(jointmanager::JointDefinition::FIXED);

	return Messages::Create(jointmanager::jointmanager,m);
}


Message
JointManager::CreateBallJoint(JointIdentifier identifier,
                              std::string const & parent,
                              std::string const & child,
                              jointmanager::JointDefinition::Frame frame,
                              const Eigen::Vector3d & position,
                              const Eigen::Vector3d & axis,
                              double drawSize){
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::Create);
	jointmanager::Create * create = m.mutable_create();

	create->set_jointidentifier(identifier);

	jointmanager::JointDefinition * definition = create->mutable_definition();

	definition->set_parent(parent);
	definition->set_child(child);
	definition->set_referenceframe(frame);

	definition->set_type(jointmanager::JointDefinition::BALL);

	definition->mutable_position()->set_x(position.x());
	definition->mutable_position()->set_y(position.y());
	definition->mutable_position()->set_z(position.z());

	definition->mutable_axis()->set_x(axis.x());
	definition->mutable_axis()->set_y(axis.y());
	definition->mutable_axis()->set_z(axis.z());

	if(drawSize > 0.0){
		definition->set_drawsize(drawSize);
	}

	return Messages::Create(jointmanager::jointmanager,m);
}
Message
JointManager::DestroyJoint(uint32_t identifier)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::Destroy);

	jointmanager::Destroy *destroy = m.mutable_destroy();

	destroy->set_jointidentifier(identifier);

	return Messages::Create(jointmanager::jointmanager,m);
}

Message JointManager::AddManagedjoint(JointIdentifier identifier,
                                      const std::string & parent,
                                      const std::string & child)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::AddManaged);
	m.mutable_addmanaged()->set_jointidentifier(identifier);
	m.mutable_addmanaged()->set_parent(parent);
	m.mutable_addmanaged()->set_child(child);

	return Messages::Create(jointmanager::jointmanager,m);
}



Message JointManager::SetJointParameter(JointIdentifier identifier,
                                        messages::jointmanager::JointParameters::Values param,
                                        double value,
                                        uint32_t dimension)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::SetJointParameter);
	m.mutable_setjointparameter()->set_jointidentifier(identifier);
	jointmanager::SetJointParameter::ParameterValue * p =
	m.mutable_setjointparameter()->add_parameters();
	p->set_parameter(param);
	p->set_value(value);
	if(dimension != 0){
		p->set_dimension(dimension);
	}
	return Messages::Create(jointmanager::jointmanager,m);

}



Message JointManager::SetJointParameters(JointIdentifier identifier,
                                         const ListOfParameterValue & listOfParameters)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::SetJointParameter);
	m.mutable_setjointparameter()->set_jointidentifier(identifier);
	m.mutable_setjointparameter()->mutable_parameters()->MergeFrom(listOfParameters);

	return Messages::Create(jointmanager::jointmanager,m);
}



Message JointManager::SetJointPassiveDynamic(JointIdentifier identifier,
                                             double stiffness,
                                             double damping,
                                             double precompression)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::SetPassiveDynamic);
	m.mutable_setpassivedynamics()->set_jointidentifier(identifier);
	m.mutable_setpassivedynamics()->mutable_dynamic()->set_stiffness(stiffness);
	m.mutable_setpassivedynamics()->mutable_dynamic()->set_damping(damping);
	m.mutable_setpassivedynamics()->mutable_dynamic()->set_precompression(precompression);

	return Messages::Create(jointmanager::jointmanager,m);
}


Message
	JointManager::SetJointPassiveDynamic(JointIdentifier identifier,
	                                     const Eigen::ArrayXd & stiffness,
	                                     const Eigen::ArrayXd & damping,
	                                     const Eigen::ArrayXd & precompression){

	unsigned int nbRows = stiffness.rows();
	typedef ::google::protobuf::RepeatedPtrField<jointmanager::PassiveDynamic> ListOfPassiveDynamic;
	if(nbRows != damping.rows() || nbRows != precompression.rows()){
		throw std::runtime_error("it should have the same size !");
	}

	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::SetPassiveDynamic);
	m.mutable_setpassivedynamics()->set_jointidentifier(identifier);
	ListOfPassiveDynamic * dynamics = m.mutable_setpassivedynamics()->mutable_multipledynamic();
	dynamics->Reserve(stiffness.rows());

	for(unsigned int i = 0; i < nbRows;++i){
		jointmanager::PassiveDynamic * d = dynamics->Add();
		d->set_stiffness(stiffness[i]);
		d->set_damping(damping[i]);
		d->set_precompression(precompression[i]);
	}
	return Messages::Create(jointmanager::jointmanager,m);
}



Message JointManager::SetJointPassiveDynamicsReferencePosition(JointIdentifier identifier,
                                                               double referencePosition)
{
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::SetPassiveDynamic);
	m.mutable_setpassivedynamics()->set_jointidentifier(identifier);
	m.mutable_setpassivedynamics()->set_refposition(referencePosition);

	return Messages::Create(jointmanager::jointmanager,m);
}


Message JointManager::AddMonitor(JointIdentifier id,
                                 jointmanager::JointDefinition::Frame frame,
                                 jointmanager::AddMonitor::Type type){
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::AddMonitor);
	m.mutable_addmonitor()->set_jointidentifier(id);
	m.mutable_addmonitor()->set_type(type);
	m.mutable_addmonitor()->set_frame(frame);

	return Messages::Create(jointmanager::jointmanager,m);
}

Message JointManager::DeleteMonitor(JointIdentifier id){
	jointmanager::JointManager m;
	m.set_action(jointmanager::Action::DeleteMonitor);
	m.mutable_deletemonitor()->set_jointidentifier(id);
	return Messages::Create(jointmanager::jointmanager,m);
}
