/**
 * \file passivedynamics.cc
 *
 *  \date Nov 24, 2011
 *  \author Alexandre Tuleu
 */

#include "passivedynamics.hh"


using namespace libwebots::plugin::modules::jointmanager;
using namespace std::tr1;

PassiveDynamics::PassiveDynamics(shared_ptr<Joint> & associatedJoint):
	d_joint(associatedJoint),
	d_referencePosition(0.0,0.0,0.0),
	d_useNegative(false)
{
}


PassiveDynamics::~PassiveDynamics()
{
}




void PassiveDynamics::SetFromMessage(const messages::jointmanager::SetPassiveDynamics & message)
{
	if(message.multipledynamic_size()){
		SetDynamicFromMessage(message.multipledynamic(),d_dynamic);
	} else if(message.has_dynamic())
	{
		SetScalarDynamicFromMessage(message.dynamic(),d_dynamic);
	}

	if(message.multiplenegativedynamic_size()){
		SetDynamicFromMessage(message.multiplenegativedynamic(),d_negativeDynamic);
		d_useNegative = true;
	} else if(message.has_negativedynamic())
	{
		SetScalarDynamicFromMessage(message.negativedynamic(),d_negativeDynamic);
		d_useNegative = true;
	}

	if(message.multiplerefposition_size()){


	} else {

	} if(message.has_refposition())
	{
		d_referencePosition = message.refposition() * Eigen::Vector3d::Ones();
	}

	if(message.has_usenegativedynamic())
	{
		d_useNegative = message.usenegativedynamic();
	}
}

#include <iostream>

void PassiveDynamics::Step() const
{
	Eigen::Array3d force = - d_dynamic.precompression.array() * d_dynamic.stiffness.array();

	Eigen::Array3d position = d_joint->GetCoordinateConst();
	Eigen::Array3d rate = d_joint->GetCoordinateRateConst();
	Eigen::Array3d diffPos = d_referencePosition - position;

	if(d_useNegative)
	{
		force+= d_negativeDynamic.stiffness * diffPos.cwiseMin(Eigen::Array3d::Zero());
		force += d_dynamic.stiffness * diffPos.cwiseMax(Eigen::Array3d::Zero()); ;
	}
	else
	{
		force += d_dynamic.stiffness * diffPos.array();
	}

	if(d_useNegative )
	{
		force -= d_negativeDynamic.damping * rate.cwiseMin(Eigen::Array3d::Zero());
		force -= d_dynamic.damping * rate.cwiseMax(Eigen::Array3d::Zero());
	}
	else
	{
		force-= rate* d_dynamic.damping.array();
	}
	d_joint->AddEffort(force.matrix());
}



void PassiveDynamics::SetScalarDynamicFromMessage(const messages::jointmanager::PassiveDynamic & message,
                                            Dynamics & dynamic)
{
	if(message.has_stiffness())
	{
		dynamic.stiffness =  message.stiffness() * Eigen::Vector3d::Ones();
	}

	if(message.has_damping())
	{
		dynamic.damping = message.damping() * Eigen::Vector3d::Ones();
	}

	if(message.has_precompression())
	{
		dynamic.precompression = message.precompression() * Eigen::Vector3d::Ones();
	}
}


void PassiveDynamics::SetDynamicFromMessage(const ListOfPassiveDynamics & data,
                                            Dynamics & dynamics){
	if(data.size() != d_joint->Dimension()){
		return;
	}
	dynamics.stiffness.setZero(3);
	dynamics.damping.setZero(3);
	dynamics.precompression.setZero(3);
	for(unsigned int i = 0; i < std::min(d_joint->Dimension(),3U); ++i){
		const messages::jointmanager::PassiveDynamic & d = data.Get(i);
		if(d.has_stiffness()){
			dynamics.stiffness[i] = d.stiffness();
		}
		if(d.has_damping()){
			dynamics.damping[i] = d.damping();
		}
		if(d.has_precompression()){
			dynamics.precompression[i] = d.precompression();
		}
	}
}

namespace mj=libwebots::plugin::modules::jointmanager;
std::ostream & mj::operator << (std::ostream & out, const PassiveDynamics & pd){
	out << " stiffness:[" << pd.d_dynamic.stiffness.transpose()
	    << "], daming:[" << pd.d_dynamic.damping.transpose()
	    << ", precompression:[" << pd.d_dynamic.precompression.transpose()
	    << "] ";

	if(pd.d_useNegative){
		out << " with histeresis : "
		    << " stiffness:[" << pd.d_negativeDynamic.stiffness.transpose()
		    << "], damping:[" << pd.d_negativeDynamic.damping.transpose()
		    << "], precompresison:[" << pd.d_negativeDynamic.precompression.transpose()
		    << "] ";
	}
	return out;
}
