/**
 * \file push.cc
 *
 *  \date Mar 2, 2011
 *  \author Alexandre Tuleu
 */

#include <plugins/physics.h>
#include <libwebots/communication/server.hh>
#include <libwebots/plugin/plugin.hh>
#include "push.hh"
#include <iostream>

using namespace libwebots::plugin::modules;

Push::PushAction::ProfileFactory Push::PushAction::s_profiles;
double Push::PushAction::s_time(0.0);

Push::Push(dWorldID world,
           dSpaceID space,
           dJointGroupID jointGroup,
           Module::MapByTypeID const &dependencies)
:
	PhysicsModule(world, space, jointGroup, dependencies),
	d_debug(GetInterface<Debug>())
{
	libwebots::communication::Server &server =
		libwebots::communication::Server::Instance();

	server.AddHandler(*this, messages::push::push, &Push::Handler);

	d_debug(messages::debug::Type::Info) << "[Push] Initialized module" << std::endl;
}

Push::~Push()
{
	for (std::list<PushAction *>::iterator push = d_pushes.begin();
	     push != d_pushes.end();
	     ++push)
	{
		delete *push;
	}
}

void
Push::RegisterDependencies()
{
	Module::RegisterDependencies();
	RegisterProfiles();
}

//
//RegisterProfiles defined in profiles.cc
//

void Push::Step()
{
	PushAction::SetTime(dWebotsGetTime());

	for (std::list<PushAction*>::iterator push = d_pushes.begin();
	     push != d_pushes.end();
	     ++push)
	{
		(*push)->Apply();
	}
}

void Push::Draw()
{
	for (std::list<PushAction*>::iterator push = d_pushes.begin();
	     push != d_pushes.end();
	     ++push)
	{
		(*push)->Draw();

		if ((*push)->Done())
		{
			// We are done, so we remove the push
			delete *push;

			push = d_pushes.erase(push);
			--push;
		}
	}
}

void Push::Handler(messages::push::Push const & message)
{
	PushAction *action = PushAction::Create(message, dWebotsGetTime(), d_debug);

	if (action)
	{
		// Creation from message was succesful
		d_pushes.push_back(action);
	}
}


Push::PushAction::PushAction(dBodyID id,
                             double x,
                             double y,
                             double z,
                             double tx,
                             double ty,
                             double tz,
                             double duration,
                             std::auto_ptr<push::Profile> profile,
                             double time)
:
	d_body(id),
	d_duration(1000.0 * duration), // webots time is in ms
	d_startTime(time),
	d_profile(profile)
{
	for (size_t i = 0; i < 6; ++i)
	{
		d_current[i] = 0.0;
	}

	d_target[0] = x;
	d_target[1] = y;
	d_target[2] = z;
	d_target[3] = tx;
	d_target[4] = ty;
	d_target[5] = tz;
}

Push::PushAction::~PushAction()
{
}

void Push::PushAction::Apply()
{
	double factor = d_profile->Value(GetRatio());

	for (size_t i = 0; i < 6; ++i)
	{
		d_current[i] = factor * d_target[i];
	}

	dBodyAddForceAtRelPos(d_body,
	                      d_current[0],
	                      d_current[1], d_current[2],
	                      0, 0, 0);
	dBodyAddTorque(d_body, d_current[3], d_current[4], d_current[5]);
}

void Push::PushAction::Draw()
{
	const dReal *pos = dBodyGetPosition(d_body);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glLineWidth(4.0);
	glBegin(GL_LINES);
	glColor3f(0.0,0.0,1.0);

	glVertex3f(pos[0], pos[1], pos[2]);
	glColor3f(0.0,1.0,1.0);
	glVertex3f(pos[0] + d_current[0] / 100,
	           pos[1] + d_current[1] / 100,
	           pos[2] + d_current[2] / 100);

	glEnd();
	glPopAttrib();
}

bool Push::PushAction::Done() const
{
	return GetRatio() >= 1.0;
}

double Push::PushAction::GetRatio() const
{
	return (s_time - d_startTime) / d_duration;

}

void Push::PushAction::SetTime(double time)
{
	s_time = time;
}

void Push::PushAction::RegisterProfile(const messages::push::Profile::Values id,
                                       push::Profile *profile)
{
	s_profiles[id] = profile;
}

Push::PushAction* Push::PushAction::Create(messages::push::Push const &message,
                                           double currentTime,
                                           interfaces::Interface<Debug> &debugOutput)
{
	dBodyID body = dWebotsGetBodyFromDEF(message.body().c_str());

	if (!body)
	{
		debugOutput() << "[Push] : No body found !" << std::endl;
		return NULL;
	}

	if (!message.has_force())
	{
		debugOutput() << "[Push] : No force found !" << std::endl;
		return NULL;
	}

	if (!message.has_duration())
	{
		debugOutput() << "[Push] : No duration found !" << std::endl;
		return NULL;
	}

	messages::push::Profile::Values profile = messages::push::Profile::Linear;

	if (message.has_profile())
	{
		profile = message.profile();
	}

	ProfileFactory::iterator profileIter = s_profiles.find(profile);

	if (profileIter == s_profiles.end())
	{
		debugOutput() << "[Push] : No profile found !" << std::endl;
		return NULL;
	}

	std::auto_ptr<push::Profile> profileToUse = (profileIter)->second->Clone();

	return new PushAction(body,
	                      message.force().x(),
	                      message.force().y(),
	                      message.force().z(),
	                      message.force().tx(),
	                      message.force().ty(),
	                      message.force().tz(),
	                      message.duration(),
	                      profileToUse,
	                      currentTime);

}
