#include "snap.hh"

#include <libwebots/communication/server.hh>
#include <libwebots/communication/messages.hh>
#include <algorithm>
#include <iostream>

#include <libwebots/plugin/utils.hh>
#include <libwebots/plugin/plugin.hh>

#include <libwebots/config.hh>

using namespace libwebots::plugin::modules;
using namespace libwebots::plugin;
using namespace std;

Snap::Snap(Module::MapByTypeID const &dependencies)
:
	Module(dependencies),
	interfaces::Debugable(GetInterface<Debug>(),"Snap")
{
	libwebots::communication::Server &server = libwebots::communication::Server::Instance();
	server.AddHandler(*this, messages::snap::snap, &Snap::Handler);

	Info() << "Initialized module" << endl;
}

void
Snap::ShortestDistance(Eigen::Vector3d const &pt,
                       Eigen::Vector3d const &point,
                       Eigen::Vector3d const &normal,
                       Eigen::Vector3d &mindist,
                       double &minlen)
{
	// Projection of vector from pt to point on the normal
	Eigen::Vector3d v = normal * (point - pt).dot(normal);
	double l = v.norm();

	if (minlen < 0 || l < minlen)
	{
		minlen = l;
		mindist = v;
	}
}

void
Snap::ShortestDistance(dGeomID geom,
                       Eigen::Vector3d const &point,
                       Eigen::Vector3d const &normal,
                       Eigen::Vector3d &mindist,
                       double &minlen)
{
	/* minx, maxx, miny, maxy, minz, maxz */
	dReal aabb[6];

	dGeomGetAABB(geom, aabb);

	ShortestDistance(Eigen::Vector3d(aabb[0], aabb[2], aabb[4]), point, normal, mindist, minlen);
	ShortestDistance(Eigen::Vector3d(aabb[0], aabb[3], aabb[4]), point, normal, mindist, minlen);
	ShortestDistance(Eigen::Vector3d(aabb[0], aabb[2], aabb[5]), point, normal, mindist, minlen);
	ShortestDistance(Eigen::Vector3d(aabb[0], aabb[3], aabb[5]), point, normal, mindist, minlen);

	ShortestDistance(Eigen::Vector3d(aabb[1], aabb[2], aabb[4]), point, normal, mindist, minlen);
	ShortestDistance(Eigen::Vector3d(aabb[1], aabb[3], aabb[4]), point, normal, mindist, minlen);
	ShortestDistance(Eigen::Vector3d(aabb[1], aabb[2], aabb[5]), point, normal, mindist, minlen);
	ShortestDistance(Eigen::Vector3d(aabb[1], aabb[3], aabb[4]), point, normal, mindist, minlen);
}

void
Snap::ShortestDistance(Body const &body,
                       Eigen::Vector3d const &point,
                       Eigen::Vector3d const &normal,
                       Eigen::Vector3d &mindist,
                       double &minlen)
{
	vector<dGeomID> geoms = body.Geoms();

	for (vector<dGeomID>::iterator iter = geoms.begin();
	     iter != geoms.end();
	     ++iter)
	{
		ShortestDistance(*iter, point, normal, mindist, minlen);
	}
}

void
Snap::Handler(messages::snap::Plane const &message)
{
	Body body = Body(message.body());

	if (!body)
	{
		Error(false) << "[Snap.Plane] The body `" << message.body() << "' could not be found..." << endl;
		return;
	}

	Eigen::Vector3d moveby(Eigen::Vector3d::Zero());
	Eigen::Vector3d point;
	Eigen::Vector3d normal;
	double movebylen = -1;

	if (message.point_size() != 3)
	{
		Error(false) << "[Snap.Plane] : Point must be a vector 3" << endl;
		return;
	}

	if (message.normal_size() != 3)
	{
		Error(false) << "[Snap.Plane] : Normal must be a vector 3" << endl;
	}

	for (size_t i = 0; i < 3; ++i)
	{
		point[i] = message.point(i);
		normal[i] = message.normal(i);
	}

	vector<Body> descendents(body.Descendents(true));

	/* Snap body (and all its children), onto the plane */
	if (message.has_closestchild() && message.closestchild())
	{
		for (vector<Body>::iterator iter = descendents.begin();
		     iter != descendents.end();
		     ++iter)
		{
			ShortestDistance(*iter, point, normal, moveby, movebylen);
		}
	}

	/* Move all descendents by moveby */
	if (movebylen >= 0)
	{
		if (message.has_margin())
		{
			moveby += normal * message.margin();
		}

		Info(false) << "[Snap.Plane] : Moving by [" << moveby.transpose() << "]" << endl;

		for (vector<Body>::iterator iter = descendents.begin();
		     iter != descendents.end();
		     ++iter)
		{
			iter->SetTranslation(iter->Translation() + moveby);
		}
	}
}

void
Snap::Handler(messages::snap::Snap const &message)
{
	switch (message.action())
	{
		case messages::snap::Action::Plane:
			if (message.has_plane())
			{
				Handler(message.plane());
			}
			else
			{
				Info() << "Message has no 'plane' field" << endl;
			}
		break;
	}
}
