#include "state.hh"
#include <libwebots/communication/server.hh>
#include <libwebots/communication/messages.hh>
#include <ode/objects.h>
#include <iostream>

#include <libwebots/plugin/plugin.hh>

#include <libwebots/plugin/utils.hh>
#include <Eigen/SVD>
#include <Eigen/LU>

using namespace libwebots::plugin::modules;
using namespace libwebots;
using namespace std;

State::State(dWorldID world,
             dSpaceID space,
             dJointGroupID jointGroup,
             Module::MapByTypeID const &dependencies)
:
	PhysicsModule(world, space, jointGroup, dependencies),
	interfaces::Debugable(GetInterface<Debug>(),"State"),
	d_previousTime(0)

{
	libwebots::communication::Server &server = libwebots::communication::Server::Instance();

	server.AddHandler(*this, messages::state::state, &State::Handler);

	Info() << "Initialized module." << endl;
}

State::~State()
{
	map<string, Monitor *>::iterator iter;

	for (iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		delete iter->second;
	}
}

void
State::StepEnd()
{
	if (d_monitors.size() == 0)
	{
		return;
	}

	communication::Server &server = communication::Server::Instance();
	double t = dWebotsGetTime() / 1000.0;
	double step = t - d_previousTime;

	for (map<string, Monitor *>::iterator iter = d_monitors.begin(); iter != d_monitors.end(); ++iter)
	{
		messages::state::Monitor monitor;

		monitor.set_name(iter->first);

		Monitor *mon = iter->second;
		dBodyID b = mon->Body;

		const dReal *pos = dBodyGetPosition(b);
		const dReal *rot = dBodyGetQuaternion(b);
		const dReal *linvel = dBodyGetLinearVel(b);
		const dReal *angvel = dBodyGetAngularVel(b);

		dVector3 angvelloc;
		dVector3 linvelloc;

		dBodyVectorFromWorld(b, angvel[0], angvel[1], angvel[2], angvelloc);
		dBodyVectorFromWorld(b, linvel[0], linvel[1], linvel[2], linvelloc);

		for (size_t i = 0; i < 3; ++i)
		{
			monitor.add_position(pos[i]);
			monitor.add_rotation(rot[i]);
			monitor.add_linear_velocity(linvelloc[i]);
			monitor.add_angular_velocity(angvelloc[i]);

			monitor.add_angular_acceleration((angvelloc[i] - mon->PreviousAngularVelocity[i]) / step);
			monitor.add_linear_acceleration((linvelloc[i] - mon->PreviousLinearVelocity[i]) / step);

			mon->PreviousAngularVelocity[i] = angvelloc[i];
			mon->PreviousLinearVelocity[i] = linvelloc[i];
		}

		monitor.add_rotation(rot[3]);
		server.Send(communication::Messages::Create(messages::state::statemonitor, monitor, iter->first));
	}

	d_previousTime = t;
}

void
State::AddMonitor(string const &name)
{
	if (d_monitors.find(name) != d_monitors.end())
	{
		return;
	}

	dBodyID body = dWebotsGetBodyFromDEF(name.c_str());

	if (body)
	{
		Monitor *monitor = new Monitor();

		monitor->Body = body;

		const dReal *linvel = dBodyGetLinearVel(body);
		const dReal *angvel = dBodyGetAngularVel(body);

		for (size_t i = 0; i < 3; ++i)
		{
			monitor->PreviousLinearVelocity[i] = linvel[i];
			monitor->PreviousAngularVelocity[i] = angvel[i];
		}

		d_monitors[name] = monitor;
	}
	else
	{
		Error() << "Could not find body for '" << name << "'" << endl;
	}
}

void
State::SetState(string const &name, messages::state::SetState const &message)
{
	dBodyID body = dWebotsGetBodyFromDEF(name.c_str());
	dGeomID geom = dWebotsGetGeomFromDEF(name.c_str());

	if (!body && !geom)
	{
		Error() << "could not find body nor geometry for '" << name << "'" << endl;
		return;
	}

	if (message.position_size() == 3)
	{
		if (body)
		{
			dBodySetPosition(body, message.position(0), message.position(1), message.position(2));
		}
		else if (geom)
		{
			dGeomSetPosition(geom, message.position(0), message.position(1), message.position(2));
		}
	}

	if (message.rotation_size() == 4)
	{
		dQuaternion q = {message.rotation(0), message.rotation(1), message.rotation(2), message.rotation(3)};

		if (body)
		{
			dBodySetQuaternion(body, q);
		}
		else if (geom && dGeomGetClass(geom) == dPlaneClass)
		{
			dVector4 n;
			dGeomPlaneGetParams(geom, n);

			dQuaternion normal = {0, n[0], n[1], n[2]};

			dQuaternion plane;
			dQMultiply0(plane, q, normal);

			dGeomPlaneSetParams(geom, plane[1], plane[2], plane[3], n[3]);
		}
		else if (geom)
		{
			dGeomSetQuaternion(geom, q);
		}
	}

	if (message.linear_velocity_size() == 3)
	{
		if (body)
		{
			dBodySetLinearVel(body, message.linear_velocity(0), message.linear_velocity(1), message.linear_velocity(2));
		}
		else if (geom)
		{
			Error() << "Cannot set linear velocity for '" << name << "' geometry." << endl;
			return;
		}
	}

	if (message.angular_velocity_size() == 3)
	{
		if (body)
		{
			dBodySetAngularVel(body, message.angular_velocity(0), message.angular_velocity(1), message.angular_velocity(2));
		}
		else if (geom)
		{
			Error() << "Cannot set angular velocity for '" << name << "' geometry." << endl;
			return;
		}
	}
}

void
State::RegisterNames(messages::state::RegisterNames const &message)
{
	for (int i = 0; i < message.names_size(); ++i)
	{
		string name = message.names(i);

		dBodyID id = dWebotsGetBodyFromDEF(name.c_str());

		if (id)
		{
			d_bodyMap[id] = name;
			d_names.push_back(name);
		}
	}

	d_bodyDump.clear();
	d_jointDump.clear();
}

void
State::Dump(messages::state::Dump &dump, dBodyID body)
{
	map<dBodyID, string>::iterator found;

	found = d_bodyMap.find(body);

	messages::state::Dump::Body &b = *(dump.add_bodies());

	if (found != d_bodyMap.end())
	{
		b.set_name(found->second);
	}
	else
	{
		b.set_name("Unknown");
	}

	dMass mass;

	dBodyGetMass(body, &mass);

	b.set_mass(mass.mass);

	for (size_t i = 0; i < 9; ++i)
	{
		b.add_inertia(mass.I[i]);
	}

	dReal const *position = dBodyGetPosition(body);
	dReal const *rotation = dBodyGetQuaternion(body);
	dReal const *linvel = dBodyGetLinearVel(body);
	dReal const *angvel = dBodyGetAngularVel(body);

	for (size_t i = 0; i < 3; ++i)
	{
		b.add_position(position[i]);
		b.add_rotation(rotation[i]);
		b.add_linear_velocity(linvel[i]);
		b.add_angular_velocity(angvel[i]);
	}

	b.add_rotation(rotation[3]);
}

dBodyID
State::CorrectedBody(dBodyID b)
{
	if (b == 0)
	{
		return b;
	}

	int num = dBodyGetNumJoints(b);

	for (int i = 0; i < num; ++i)
	{
		dJointID j = dBodyGetJoint(b, i);

		if (dJointGetType(j) == dJointTypeFixed)
		{
			if (dJointGetBody(j, 1) == b)
			{
				return dJointGetBody(j, 0);
			}
		}
	}

	return b;
}

void
State::Dump(messages::state::Dump &dump, dJointID joint, map<dBodyID, size_t> const &bodyIndex)
{
	dBodyID b1 = CorrectedBody(dJointGetBody(joint, 0));
	dBodyID b2 = CorrectedBody(dJointGetBody(joint, 1));

	// If one of the bodies has a fixed joint, then just consider the other
	// body as the one thingie thingie

	if (bodyIndex.find(b1) == bodyIndex.end() || bodyIndex.find(b2) == bodyIndex.end())
	{
		return;
	}

	dVector3 anchor = {0, 0, 0};
	dVector3 axis = {0, 0, 0};

	switch (dJointGetType(joint))
	{
		case dJointTypeHinge:
			dJointGetHingeAnchor(joint, anchor);
			dJointGetHingeAxis(joint, axis);
		break;
		case dJointTypeSlider:
			dJointGetSliderAxis(joint, axis);
		break;
		case dJointTypeNone:
		case dJointTypeBall:
		case dJointTypeContact:
		case dJointTypeUniversal:
		case dJointTypeHinge2:
		case dJointTypeFixed:
		case dJointTypeNull:
		case dJointTypeAMotor:
		case dJointTypeLMotor:
		case dJointTypePlane2D:
		case dJointTypePR:
		case dJointTypePU:
		case dJointTypePiston:
		default:
			return;
		break;
	}

	messages::state::Dump::Joint &j = *(dump.add_joints());

	j.set_body1(bodyIndex.find(b1)->second);
	j.set_body2(bodyIndex.find(b2)->second);

	for (size_t i = 0; i < 3; ++i)
	{
		j.add_axis(axis[i]);
		j.add_anchor(anchor[i]);
	}
}

void
State::Dump(vector<dBodyID> const &bodies, vector<dJointID> const &joints)
{
	messages::state::Dump dump;

	map<dBodyID, size_t> bodyIndex;
	size_t i = 0;

	for (vector<dBodyID>::const_iterator iter = bodies.begin(); iter != bodies.end(); ++iter)
	{
		Dump(dump, *iter);

		bodyIndex[*iter] = i++;
	}

	for (vector<dJointID>::const_iterator iter = joints.begin(); iter != joints.end(); ++iter)
	{
		Dump(dump, *iter, bodyIndex);
	}

	communication::Server &server = communication::Server::Instance();
	server.Send(communication::Messages::Create(messages::state::statedump, dump));
}

void
State::Dump(string const &body)
{
	dBodyID origin = dWebotsGetBodyFromDEF(body.c_str());

	if (!origin)
	{
		return;
	}

	vector<dBodyID> bodies = Utils::AllBodies(origin);
	vector<dJointID> joints = Utils::AllJoints(origin);

	Dump(bodies, joints);
}

void
State::InitializeDump()
{
	// Make list of all bodies and joints from registered names
	d_bodyDump.clear();
	d_jointDump.clear();

	map<dBodyID, bool> doneb;

	for (vector<string>::iterator iter = d_names.begin(); iter != d_names.end(); ++iter)
	{
		dBodyID origin = dWebotsGetBodyFromDEF(iter->c_str());

		if (doneb.find(origin) != doneb.end())
		{
			continue;
		}

		doneb[origin] = true;
		d_bodyDump.push_back(origin);
	}

	for (vector<string>::iterator iter = d_names.begin(); iter != d_names.end(); ++iter)
	{
		dBodyID origin = dWebotsGetBodyFromDEF(iter->c_str());
		vector<dBodyID> subsetb = Utils::AllBodies(origin);

		for (vector<dBodyID>::iterator it = subsetb.begin(); it != subsetb.end(); ++it)
		{
			if (doneb.find(*it) == doneb.end())
			{
				doneb[*it] = true;
				d_bodyDump.push_back(*it);
			}
		}
	}

	map<dJointID, bool> donej;

	for (vector<dBodyID>::iterator iter = d_bodyDump.begin(); iter != d_bodyDump.end(); ++iter)
	{
		vector<dJointID> subsetj = Utils::AllJoints(*iter);

		for (vector<dJointID>::iterator it = subsetj.begin(); it != subsetj.end(); ++it)
		{
			if (donej.find(*it) == donej.end())
			{
				donej[*it] = true;
				d_jointDump.push_back(*it);
			}
		}
	}
}

void
State::Dump()
{
	if (d_bodyDump.empty())
	{
		InitializeDump();
	}

	Dump(d_bodyDump, d_jointDump);
}

vector<plugin::Body>
State::FindStateTrees(messages::state::InitialState const &message,
                      map<dBodyID, messages::state::InitialState::Description> &defs)
{
	vector<plugin::Body> trees;

	for (int i = 0; i < message.joints_size(); ++i)
	{
		messages::state::InitialState::Description const &desc = message.joints(i);

		// Lookup body
		dBodyID id = dWebotsGetBodyFromDEF(desc.name().c_str());

		if (!id)
		{
			Error(false) << "[State.InitialState] : Could not find body "
			             << "`"
			             << desc.name()
			             << "'"
			             << endl;
			continue;
		}

		dBodyID parentId = 0;
		Utils::FindWebotsJoint(id, parentId);

		if (!parentId)
		{
			parentId = id;
		}

		defs[id] = desc;
		bool found = false;

		/* Generate a new tree for this body */
		for (vector<plugin::Body>::iterator iter = trees.begin(); iter != trees.end(); ++iter)
		{
			if (iter->Includes(parentId))
			{
				found = true;
				break;
			}
		}

		if (found)
		{
			continue;
		}

		plugin::Body tree = plugin::Body(parentId);

		vector<plugin::Body>::iterator iter = trees.begin();

		while (iter != trees.end())
		{
			if (tree.Includes(*iter))
			{
				iter = trees.erase(iter);
			}
			else
			{
				++iter;
			}
		}

		trees.push_back(tree);
	}

	return trees;
}

void
State::InitialState(plugin::Body &parent,
                    map<dBodyID, messages::state::InitialState::Description> const &defs,
                    Eigen::Isometry3d const &transform)
{
	// Calculate for the chain of bodies the angle, ang vel and lin vel
	Eigen::Vector3d wp = parent.AngularVelocity();
	Eigen::Vector3d vp = parent.LinearVelocity();
	Eigen::Vector3d const &pp = parent.Translation();

	map<dBodyID, messages::state::InitialState::Description>::const_iterator found;
	map<dJointID, plugin::Body>::iterator iter;

	found = defs.find(parent.Id());

	// Set linear velocity on parent if it has a definition
	if (found != defs.end())
	{
		Eigen::Vector3d linvel = vp;
		messages::state::InitialState::Description const &def = found->second;

		if (def.linearvelocity_size() == 3)
		{
			for (size_t i = 0; i < 3; ++i)
			{
				linvel[i] += def.linearvelocity(i);
			}
		}

		vp = linvel;
		parent.SetLinearVelocity(vp);
	}

	// Iterate over all the children to calculate angular velocity and new
	// initial velocity
	for (iter = parent.begin(); iter != parent.end(); ++iter)
	{
		plugin::Body &child = iter->second;
		dJointID joint = iter->first;
		dVector3 odeanchor;
		dVector3 odeaxis;

		found = defs.find(child.Id());

		// Set the position, rotation and axis according to the current
		// transformation
		Eigen::Isometry3d lc = transform;
		Eigen::Vector3d anchor(pp);
		Eigen::Vector3d axis(Eigen::Vector3d::Zero());

		if (dJointGetType(joint) == dJointTypeHinge)
		{
			dJointGetHingeAxis(joint, odeaxis);
			dJointGetHingeAnchor(joint, odeanchor);

			axis = Eigen::Vector3d(odeaxis);
			anchor = Eigen::Vector3d(odeanchor);

			if (found != defs.end())
			{
				// Then rotate according to axis and angle
				messages::state::InitialState::Description const &def = found->second;
				lc.rotate(Eigen::AngleAxis<double>(def.angle(), axis));
			}
		}

		child.SetRotation(child.Rotation() * lc.rotation());
		child.SetTranslation(anchor + lc.rotation() * child.LocalTranslation());

		Eigen::Vector3d angvel(Eigen::Vector3d::Zero());
		Eigen::Vector3d linvel(Eigen::Vector3d::Zero());

		if (found != defs.end())
		{
			messages::state::InitialState::Description const &def = found->second;

			angvel = axis * def.angularvelocity();

			if (def.linearvelocity_size() == 3)
			{
				for (size_t i = 0; i < 3; ++i)
				{
					linvel[i] = def.linearvelocity(i);
				}
			}
		}

		// Angular velocity adds up
		child.SetAngularVelocity(wp + angvel);

		Eigen::Vector3d parentToAnchor = anchor - pp;
		Eigen::Vector3d anchorToChild = child.Translation() - anchor;

		Eigen::Vector3d rest = wp.cross(parentToAnchor);
		Eigen::Vector3d vvc = child.AngularVelocity().cross(anchorToChild);

		child.SetLinearVelocity(vp + rest + vvc);
		InitialState(child, defs, lc);
	}
}

void
State::InitialState(messages::state::InitialState const &message)
{
	// Recursively initialize the initial angle and the velocity of all the
	// specified joints, fun
	map<dBodyID, messages::state::InitialState::Description> defs;
	vector<plugin::Body> trees;

	// Find all independent trees for which some of the bodies have an
	// initial state
	trees = FindStateTrees(message, defs);

	// Set initial angle, angular velocity and linear velocity on each
	// of the independent trees
	for (vector<plugin::Body>::iterator iter = trees.begin(); iter != trees.end(); ++iter)
	{
		Eigen::Isometry3d transform(Eigen::Isometry3d::Identity());

		InitialState(*iter, defs, transform);
	}
}

void
State::Handler(messages::state::State const &message)
{
	switch (message.action())
	{
		case messages::state::Action::Push:
			d_state.Push();
		break;
		case messages::state::Action::Pop:
			d_state.Pop();
		break;
		case messages::state::Action::Apply:
			d_state.Apply();
		break;
		case messages::state::Action::Update:
			d_state.Update();
		break;
		case messages::state::Action::RegisterBody:
			if (message.has_body())
			{
				d_state.Register(message.body());
			}
			else
			{
				Error() << "Missing 'body' field" << endl;
			}
		break;
		case messages::state::Action::MonitorBody:
			if (message.has_body())
			{
				AddMonitor(message.body());
			}
			else
			{
				Error() << "Missing 'body' field" << endl;
			}
		break;
		case messages::state::Action::Set:
			if (message.has_body() && message.has_setstate())
			{
				SetState(message.body(), message.setstate());
			}
			else if (!message.has_body())
			{
				Error() << "Missing 'body' field" << endl;
			}
			else
			{
				Error() << "Missing 'setstate' field" << endl;
			}
		break;
		case messages::state::Action::RegisterNames:
			if (!message.has_registernames())
			{
				Error() << "Missing 'registernames' field" << endl;
			}
			else
			{
				RegisterNames(message.registernames());
			}
		break;
		case messages::state::Action::Dump:
			if (message.has_body())
			{
				Dump(message.body());
			}
			else
			{
				Dump();
			}
		break;
		case messages::state::Action::InitialState:
			InitialState(message.initialstate());
		break;
	}
}
