#include "plugin.hh"

#include <vector>
#include <iostream>
#include <dlfcn.h>
#include <unistd.h>

#include <libwebots/communication/server.hh>

using namespace std;
using namespace libwebots::plugin;
using namespace libwebots::communication;

#if defined(APPLE) || defined(MACOS)
#define get_current_dir_name() (getcwd(NULL,0))
#endif

double Plugin::s_previousWebotsTime(0.0);
double Plugin::s_timestep(0.0);

Plugin *Plugin::s_instance = 0;

/**
 * @class libwebots::plugin::Plugin
 * @brief Main webots plugin interface
 *
 * The Plugin class is the main interface for handling modules and handling
 * webots callbacks to be routed on each of the registered PhysicsModule.
 *
 * @par Example
 * @code
 * void
 * webots_plugin_register_modules()
 * {
 *   // Register the State module
 *   Plugin::Instance().Register<modules::State>();
 * }
 * @endcode
 *
 **/

/**
 * @brief Initialize the plugin
 *
 * Initialize the singleton plugin object.
 *
 * @returns the singleton plugin object.
 *
 **/
Plugin &
Plugin::Initialize(dWorldID world, dSpaceID space, dJointGroupID jointGroup)
{
	if (s_instance)
	{
		delete s_instance;
	}

	s_instance = new Plugin(world, space, jointGroup);

	return *s_instance;
}

double Plugin::TimeStep(){
	return s_timestep;
}

/**
 * @brief Singleton instance accessor
 *
 * Get the plugin singleton instance.
 *
 * @returns the singleton plugin object.
 *
 **/
Plugin &
Plugin::Instance()
{
	return *s_instance;
}

Plugin::~Plugin()
{
	for (Module::MapByTypeID::iterator iter = d_allModules.begin();
	     iter != d_allModules.end();
	     ++iter)
	{
		delete iter->second;
	}

	d_allModules.clear();
	d_physicsModules.clear();

	Server::Finalize();
}

Plugin::Plugin(dWorldID world, dSpaceID space, dJointGroupID jointGroup)
:
	WebotsInterface(world, space, jointGroup)
{
	Server::Initialize();

	char *path = get_current_dir_name();
	d_path = string(path);
	free(path);
}

string const &
Plugin::Path() const
{
	return d_path;
}

void
Plugin::ProcessMessages()
{
	Server &server = Server::Instance();
	server.Accept();

	Client *client;
	Message message;

	while (server.Receive(client, message))
	{
		// Drop all messages that do not have handlers
	}
}

/**
 * @brief Step method
 *
 * Overridden to call Step() on each of the registered PhysicsModule. Before
 * calling routing Step(), incoming messages will be processed.
 *
 **/
void
Plugin::Step()
{

	if(dWebotsGetTime() == 0.0){
		s_timestep           = 0.0;
		s_previousWebotsTime = 0.0;
	}

	if(s_previousWebotsTime == 0.0){
		s_timestep = dWebotsGetTime() - s_previousWebotsTime;
		s_timestep /= 1e3;
		s_previousWebotsTime = dWebotsGetTime();
	}

	ProcessMessages();

	for (PhysicsModule::List::iterator iter = d_physicsModules.begin();
	     iter != d_physicsModules.end();
	     ++iter)
	{
		(*iter)->Step();
	}
}

/**
 * @brief StepEnd method
 *
 * Overridden to call StepEnd() on each of the registered PhysicsModule.
 *
 **/
void
Plugin::StepEnd()
{
	for (PhysicsModule::List::iterator iter = d_physicsModules.begin();
	     iter != d_physicsModules.end();
	     ++iter)
	{
		(*iter)->StepEnd();
	}
}

/**
 * @brief Collide method
 * @param g1 first geom
 * @param g2 second geom
 *
 * Overridden to call Collide() on each of the registered PhysicsModule.
 *
 * @note The first registered module that handles the collision will return
 * this call and prevent other modules to handle any collisions (in order
 * of registration).
 *
 * @returns true if the collision was handled, false otherwise
 **/
bool
Plugin::Collide(dGeomID g1, dGeomID g2)
{
	for (PhysicsModule::List::iterator iter = d_physicsModules.begin();
			iter != d_physicsModules.end(); ++iter)
	{
		if ((*iter)->Collide(g1, g2))
		{
			return true;
		}
	}

	return false;
}

/**
 * @brief Draw method
 *
 * Overridden to call Draw() on each of the registered PhysicsModule.
 *
 **/
void
Plugin::Draw()
{
	for (PhysicsModule::List::iterator iter = d_physicsModules.begin();
	     iter != d_physicsModules.end();
	     ++iter)
	{
		(*iter)->Draw();
	}
}

/**
 * @brief PreDraw method
 *
 * Overridden to call PreDraw() on each of the registered PhysicsModule.
 *
 **/
void
Plugin::PreDraw()
{
	for (PhysicsModule::List::iterator iter = d_physicsModules.begin();
	     iter != d_physicsModules.end();
	     ++iter)
	{
		(*iter)->PreDraw();
	}
}

/**
 * @brief Finalize singleton instance
 *
 * Finalize the plugin singleton instance
 *
 **/
void
Plugin::Finalize()
{
	if (s_instance)
	{
		delete s_instance;
		s_instance = 0;
	}
}


Plugin::DependencyStack::DependencyStack()
{
}

Plugin::DependencyStack::~DependencyStack()
{
}

void
Plugin::DependencyStack::Push()
{
	d_stackOfDependencies.push_back(Module::MapByTypeID());
}

void
Plugin::DependencyStack::Pop()
{
	d_stackOfDependencies.pop_back();
}

void
Plugin::DependencyStack::Add(TypeID type, Module *module)
{
	if (d_stackOfDependencies.empty())
	{
		return;
	}

	d_stackOfDependencies.back().insert(std::make_pair(type, module));
}

Module::MapByTypeID const &
Plugin::DependencyStack::Data() const
{
	if (d_stackOfDependencies.empty())
	{
		return s_void;
	}

	return d_stackOfDependencies.back();
}

Module::MapByTypeID const Plugin::DependencyStack::s_void;
