#ifndef __LIBWEBOTS_PLUGIN_PLUGIN_H__
#define __LIBWEBOTS_PLUGIN_PLUGIN_H__

#include <plugins/physics.h>
#include <libwebots/plugin/definitions.h>
#include <libwebots/plugin/baseof.hh>
#include <libwebots/plugin/physicsmodule.hh>
#include <list>
#include "typemap.hh"
#include <iostream>
#include <exception>

namespace libwebots
{
namespace plugin
{
	class Plugin : public WebotsInterface
	{
		std::string d_path;

		public:
			static double TimeStep();
			static Plugin &Initialize(dWorldID world,
			                          dSpaceID space,
			                          dJointGroupID jointGroup);
			~Plugin();

			static Plugin &Instance();
			static void Finalize();

			std::string const &Path() const;

			virtual void Draw();
			virtual void PreDraw();
			virtual void Step();
			virtual void StepEnd();
			virtual bool Collide(dGeomID g1, dGeomID g2);

			template <typename T>
			void Register();
		private:
			class DependencyStack
			{
				public:
					DependencyStack();
					~DependencyStack();

					void Push();

					void Pop();
					Module::MapByTypeID const &Data() const;
					void Add(TypeID type, Module *module);
				private:
					const static Module::MapByTypeID s_void;
					std::list<Module::MapByTypeID> d_stackOfDependencies;
			};

			template <class ModuleName, bool isPhysicsPlugin>
			class ModuleCreator
			{
				public :
					static Module *Create(dWorldID ,
					                      dSpaceID ,
					                      dJointGroupID ,
					                      Module::MapByTypeID const &dependencies)
					{
						return new ModuleName(dependencies);
					}
			};

			template <class ModuleName>
			class ModuleCreator<ModuleName, true>
			{
				public :
					static Module *Create(dWorldID world,
					                      dSpaceID space,
					                      dJointGroupID group,
					                      Module::MapByTypeID const &dependencies)
					{
						return new ModuleName(world, space, group, dependencies);
					}
			};

			template <typename T, size_t CanCall>
			struct Caller
			{
				static void Call()
				{
				}
			};

			template <typename T>
			struct Caller<T, 1>
			{
				static void Call()
				{
					T::RegisterDependencies();
				}
			};

			template <typename T>
			class CanRegisterDependencies
			{
				typedef char HasRegisterDependencies;
				typedef long LacksRegisterDependencies;

				template <typename C>
				static HasRegisterDependencies Test(typeof(&C::RegisterDependencies));

				template <typename C>
				static LacksRegisterDependencies Test(...);

				public:
					enum
					{
						Value = sizeof(Test<T>(0)) == sizeof(char)
					};

					static void RegisterDependencies()
					{
						Caller<T, Value>::Call();
					}
			};

			void ProcessMessages();

			template <typename T>
			bool IsAPhysicsModule();

			Plugin(dWorldID world, dSpaceID space, dJointGroupID jointGroup);

			static Plugin *s_instance;
			static double s_previousWebotsTime;
			static double s_timestep;

			Module::MapByTypeID d_allModules;
			PhysicsModule::List d_physicsModules;
			DependencyStack d_dependenciesStack;
	};

	/**
	 * @brief Register a module
	 *
	 * This is the main method of the plugin that is used by modules
	 * (and by webots_plugin_register_modules()). This method registers a
	 * module type to be activated. This can be used in the
	 * webots_plugin_register_modules() callback implemented by Webots
	 * physics plugins to register all the modules that should be
	 * activated. In addition, this method can be used by a module in
	 * its Module::RegisterDependencies() static method to register
	 * dependencies.
	 *
	 **/
	template <typename T>
	void
	Plugin::Register()
	{
		TypeID typeToRegister = TypeMap<T>::ID();
		Module::MapByTypeID::const_iterator iter = d_allModules.find(typeToRegister);

		if (iter != d_allModules.end())
		{
			if (!iter->second)
			{
				//WTF, CYCLIC DEPENDENCY ABORD ABORD ABORD
				std::cerr << "You have a cyclic dependency this is really bad, you "
				          << "will be punished for that, now you can restart webots!"
				          << std::endl;

				int *cyclicDependency = 0;
				*cyclicDependency = 42;
			}

			//still have to put the dependency on the stack.
			d_dependenciesStack.Add(typeToRegister, iter->second);

			return;
		}

		d_allModules[typeToRegister] = 0;

		d_dependenciesStack.Push();

		CanRegisterDependencies<T>::RegisterDependencies();

		Module *module =
				ModuleCreator<T, is_base_of<PhysicsModule, T>::value>::Create(World(),
				                                                              Space(),
				                                                              JointGroup(),
				                                                              d_dependenciesStack.Data());

		if (is_base_of<PhysicsModule, T>::value)
		{
			d_physicsModules.push_back(dynamic_cast<PhysicsModule *>(module));
		}

		d_dependenciesStack.Pop();

		d_dependenciesStack.Add(typeToRegister, module);
		d_allModules[typeToRegister] = module;
	}
}
}

#endif /* __LIBWEBOTS_PLUGIN_PLUGIN_H__ */

