#include "state.hh"

#include <iostream>

using namespace std;
using namespace libwebots::plugin;

State::Item::Item(dBodyID body)
:
	Body(body)
{
	Store();
}

void
State::Item::Store()
{
	const dReal *position = dBodyGetPosition(Body);
	const dReal *linvel = dBodyGetLinearVel(Body);
	const dReal *angvel = dBodyGetAngularVel(Body);
	const dReal *rotation = dBodyGetQuaternion(Body);

	for (size_t i = 0; i < 3; ++i)
	{
		Position[i] = position[i];
		LinearVelocity[i] = linvel[i];
		AngularVelocity[i] = angvel[i];

		Rotation[i] = rotation[i];
	}

	Rotation[3] = rotation[3];
}

void
State::Item::Reset()
{
	if (!Body)
	{
		return;
	}

	dBodySetPosition(Body, Position[0], Position[1], Position[2]);
	dBodySetQuaternion(Body, Rotation);
	dBodySetLinearVel(Body, LinearVelocity[0], LinearVelocity[1], LinearVelocity[2]);
	dBodySetAngularVel(Body, AngularVelocity[0], AngularVelocity[1], AngularVelocity[2]);
}

State::State()
{
}

State::State(dBodyID root)
{
	Register(root);
}

State::State(dSpaceID space)
{
	Register(space);
}

void
State::Register(dBodyID root)
{
	ResolveJoints(root);
}

void
State::Register(string const &bodyDef)
{
	Register(dWebotsGetBodyFromDEF(bodyDef.c_str()));
}

void
State::Register(dSpaceID space)
{
	for (int i = 0; i < dSpaceGetNumGeoms(space); ++i)
	{
		dBodyID body = dGeomGetBody(dSpaceGetGeom(space, i));
		Register(body);
	}
}

bool
State::BodyExists(dBodyID body) const
{
	for (vector<dBodyID>::const_iterator iter = d_bodies.begin(); iter != d_bodies.end(); ++iter)
	{
		if (*iter == body)
		{
			return true;
		}
	}

	return false;
}

void
State::ResolveJoints(dBodyID body)
{
	if (!body)
	{
		return;
	}

	int num = dBodyGetNumJoints(body);

	d_bodies.push_back(body);

	for (int i = 0; i < num; ++i)
	{
		dJointID joint = dBodyGetJoint(body, i);
		dBodyID other = dJointGetBody(joint, 1);

		if (other == body)
		{
			other = dJointGetBody(joint, 0);
		}

		if (!other || BodyExists(other))
		{
			continue;
		}

		d_bodies.push_back(other);
		ResolveJoints(other);
	}
}

void
State::Push()
{
	vector<Item *> state;
	state.resize(d_bodies.size());

	for (size_t i = 0; i < d_bodies.size(); ++i)
	{
		state[i] = new Item(d_bodies[i]);
	}

	d_state.push(state);
}

void
State::Update()
{
	if (d_state.empty())
	{
		Push();
		return;
	}

	vector<Item *> &items = d_state.top();

	for (vector<Item *>::iterator iter = items.begin(); iter != items.end(); ++iter)
	{
		(*iter)->Store();
	}
}

void
State::Apply()
{
	if (d_state.empty())
	{
		return;
	}

	vector<Item *> &items = d_state.top();

	for (vector<Item *>::iterator iter = items.begin(); iter != items.end(); ++iter)
	{
		(*iter)->Reset();
	}
}

void
State::Pop()
{
	d_state.pop();
}

vector<dBodyID>
State::Bodies()
{
	return d_bodies;
}

vector<dBodyID> const &
State::Bodies() const
{
	return d_bodies;
}

vector<State::Item *>
State::Top()
{
	return d_state.top();
}

vector<State::Item *> const &
State::Top() const
{
	return d_state.top();
}
