/**
 * \file webotsinterface.cc
 *
 *  \date Mar 7, 2011
 *  \author Alexandre Tuleu
 */

#include "webotsinterface.hh"

using namespace libwebots::plugin;

/**
 * @class libwebots::plugin::WebotsInterface
 * @brief Webots plugin interface
 *
 * A base class providing virtual members for Webots plugin hooks
 * (e.g. Step(), StepEnd(), Collide(), etc.). This interfaces maps directly
 * to the webots plugin callback interface.
 *
 * @see
 * @li <a href="http://www.cyberbotics.com/reference/chapter6.php">Webots Reference Manual, Physics Plugin</a>
 * @li <a href="http://www.cyberbotics.com/reference/section6.3.php">Webots Reference Manual, Physics Plugin Callbacks</a>
 *
 **/

/**
 * @brief Create new object
 *
 * Create a new object.
 *
 **/
WebotsInterface::WebotsInterface(dWorldID world,
                                 dSpaceID space,
                                 dJointGroupID jointGroup)
:
	d_world(world),
	d_space(space),
	d_jointGroup(jointGroup)
{
}

/**
 * @brief Destructor
 *
 * Destructor
 *
 **/
WebotsInterface::~WebotsInterface()
{
}

/**
 * @brief Step
 *
 * Called just before the next ODE physics integration. Use this method
 * to add additional torques and forces to particular bodies.
 *
 **/
void
WebotsInterface::Step()
{
}

/**
 * @brief StepEnd
 *
 * Called just after an ODE physics integration. Use this method to read back
 * force feedback (for example from collision contact joints).
 *
 **/
void
WebotsInterface::StepEnd()
{
}

/**
 * @brief Collide
 * @param g1 the first geom
 * @param g2 the second geom
 *
 * Called to collide to geoms. Use this method to implement a custom collision
 * detection.
 *
 * @returns true if the collision was handled, or false if the collision should
 * be handled by someone else
 *
 **/
bool
WebotsInterface::Collide(dGeomID , dGeomID )
{
	return false;
}

/**
 * @brief PreDraw
 *
 * Called just before Webots draws the scene
 *
 **/
void
WebotsInterface::PreDraw()
{
}

/**
 * @brief Draw
 *
 * Called just after Webots has drawn the scene
 *
 **/
void
WebotsInterface::Draw()
{
}

/**
 * Get world id
 *
 * Get Webots ODE world id
 *
 **/
dWorldID
WebotsInterface::World() const
{
	return d_world;
}

/**
 * @brief Get space id
 *
 * Get Webots ODE space id
 *
 **/
dSpaceID
WebotsInterface::Space() const
{
	return d_space;
}

/**
 * @brief Get joint group id
 *
 * Get Webots ODE contact joint group id
 *
 **/
dJointGroupID
WebotsInterface::JointGroup() const
{
	return d_jointGroup;
}
