#include "commandline.hh"

#include <map>
#include <unistd.h>
#include <getopt.h>
#include <exception>
#include <typeinfo>
#include <iostream>
#include <sstream>
#include <iomanip>

using namespace libwebots::tools::protodoc;
using namespace std;

CommandLine *CommandLine::s_instance(0);

CommandLine &
CommandLine::Initialize(string const &name, string const &rest)
{
	if (s_instance)
	{
		delete s_instance;
	}

	s_instance = new CommandLine(name, rest);
	return *s_instance;
}

CommandLine &
CommandLine::Instance()
{
	if (!s_instance)
	{
		s_instance = new CommandLine();
	}

	return *s_instance;
}

CommandLine::CommandLine(string const &name,
                         string const &rest)
:
	d_name(name),
	d_rest(rest)
{
}

CommandLine::~CommandLine()
{
	for (vector<Option *>::iterator iter = d_options.begin(); iter != d_options.end(); ++iter)
	{
		delete *iter;
	}
}

CommandLine::Option::~Option()
{
}

CommandLine::Option::Option(string const &name,
                            char abbrev,
                            string const &description,
                            bool defaultValue)
:
	Name(name),
	Abbrev(abbrev),
	Description(description),
	Value(defaultValue)
{
}

CommandLine::Option::operator bool() const
{
	return Value;
}

CommandLine::OptionArg::OptionArg(string const &name,
                                  char abbrev,
                                  string const &description,
                                  string const &defaultValue)
:
	Option(name, abbrev, description, false),
	Argument(defaultValue)
{
}

void
CommandLine::Parse(int argc, char * const *argv)
{
	// Build option string
	vector<Option *>::iterator iter;
	string optstr;
	struct option *longopts = new struct option[d_numArgs];
	size_t argptr = 0;

	map<int, Option *> shortmap;
	vector<Option *> longmap;

	for (iter = d_options.begin(); iter != d_options.end(); ++iter)
	{
		Option &opt = **iter;
		OptionArg *arg = 0;

		try
		{
			arg = dynamic_cast<OptionArg *>(*iter);
		}
		catch (bad_cast)
		{
		}

		if (opt.Abbrev)
		{
			optstr.push_back(opt.Abbrev);

			if (arg)
			{
				optstr.push_back(':');
			}

			shortmap[static_cast<int>(opt.Abbrev)] = *iter;
		}

		if (opt.Name != "")
		{
			longopts[argptr].name = opt.Name.c_str();
			longopts[argptr].has_arg = arg != 0;
			longopts[argptr].flag = reinterpret_cast<int *>(&(opt.Value));
			longopts[argptr].val = !(opt.Value);

			longmap.push_back(*iter);

			++argptr;
		}
	}

	int longopt = -1;
	int val;

	while ((val = getopt_long(argc, argv, optstr.c_str(), longopts, &longopt)) != -1)
	{
		Option *opt;

		if (longopt != -1)
		{
			opt = longmap[longopt];
		}
		else
		{
			opt = shortmap[val];
			opt->Value = !opt->Value;
		}

		OptionArg *arg = 0;

		try
		{
			arg = dynamic_cast<OptionArg *>(opt);
		}
		catch (bad_cast)
		{
		}

		if (arg)
		{
			arg->Argument = optarg;
			arg->Arguments.push_back(optarg);
		}

		longopt = -1;
	}

	for (int i = optind; i < argc; ++i)
	{
		d_arguments.push_back(argv[i]);
	}

	delete[] longopts;
}

void
CommandLine::ShowHelp()
{
	cerr << "Usage:" << endl << "  " << d_name;

	if (d_options.size() > 0)
	{
		cerr << " [OPTION...]";
	}

	if (!d_rest.empty())
	{
		cerr << " " << d_rest;
	}

	cerr << endl << endl;

	if (d_options.empty())
	{
		return;
	}

	cerr << "Options:" << endl;

	vector<Option *>::const_iterator iter;
	vector<pair<string, string> > opts;
	size_t maxlen = 0;

	for (iter = d_options.begin(); iter != d_options.end(); ++iter)
	{
		Option &opt = **iter;
		OptionArg *arg = 0;

		stringstream ss;

		try
		{
			arg = dynamic_cast<OptionArg *>(*iter);
		}
		catch (bad_cast)
		{
		}

		if (opt.Abbrev)
		{
			ss << "-" << opt.Abbrev;

			if (opt.Name.empty() && arg)
			{
				ss << "=VAL";
			}
			else if (!opt.Name.empty())
			{
				ss << ", ";
			}
		}
		else
		{
			ss << "    ";
		}

		if (!opt.Name.empty())
		{
			ss << "--" << opt.Name;

			if (arg)
			{
				ss << "=VAL";
			}
		}

		string s = ss.str();
		size_t len = s.size();

		opts.push_back(pair<string, string>(s, opt.Description));

		if (len > maxlen)
		{
			maxlen = len;
		}
	}

	vector<pair<string, string> >::iterator oiter;

	for (oiter = opts.begin(); oiter != opts.end(); ++oiter)
	{
		cerr << "  " << setw(maxlen) << left << oiter->first << "    " << oiter->second << endl;
	}

	cerr << endl;
}

vector<string> const &
CommandLine::Arguments() const
{
	return d_arguments;
}
