#include "docfinder.hh"
#include <fstream>
#include <iostream>
#include <algorithm>
#include <sstream>

using namespace libwebots::tools::protodoc;
using namespace std;
using namespace google::protobuf;

DocFinder::DocFinder()
{
}

DocFinder::DocFinder(string const &filename,
                     FileDescriptorProto const &proto)
:
	d_filename(filename)
{
	Scan(proto);
}

bool
DocFinder::Lookup(Message const &message, string &comment)
{
	map<Message const *, Documentation>::iterator iter;

	iter = d_locations.find(&message);

	if (iter != d_locations.end())
	{
		comment = iter->second.Comment;
		return true;
	}

	return false;
}

string
DocFinder::Lookup(Message const &message)
{
	string ret;

	if (!Lookup(message, ret))
	{
		return "";
	}
	else
	{
		return ret;
	}
}

bool
DocFinder::LocationMatch(SourceCodeInfo::Location const &location,
                         deque<int>               const &path)
{
	if (path.size() != location.path_size())
	{
		return false;
	}

	for (int i = 0; i < path.size(); ++i)
	{
		if (path[path.size() - i - 1] != location.path(i))
		{
			return false;
		}
	}

	return true;
}

bool
DocFinder::FindPath(FileDescriptorProto const &proto,
                    deque<int>          const &path,
                    SourceCodeInfo::Location  &location)
{
	SourceCodeInfo const &info = proto.source_code_info();

	for (int i = 0; i < info.location_size(); ++i)
	{
		SourceCodeInfo::Location const &loc = info.location(i);

		if (LocationMatch(loc, path))
		{
			location = loc;
			return true;
		}
	}

	return false;
}

string
DocFinder::ExtractDocumentation(SourceCodeInfo::Location const &location)
{
	int line = location.span(0);
	int column = location.span(1);
	bool incomment = false;
	stringstream ret;

	if (line < 0 || d_lines.size() <= line)
	{
		return "";
	}

	while (line >= 0)
	{
		--column;

		if (column < 0)
		{
			--line;

			if (line >= 0)
			{
				column = d_lines[line].size();

				if (incomment)
				{
					ret << endl;
				}
			}

			continue;
		}

		if (incomment)
		{
			if (d_lines[line][column] == '*' && column > 0 && d_lines[line][column - 1] == '/')
			{
				string s = ret.str();
				reverse(s.begin(), s.end());

				return s;
			}
			else
			{
				ret << d_lines[line][column];
			}
		}
		else if (d_lines[line][column] == '/' && column > 0 && d_lines[line][column - 1] == '*')
		{
			--column;
			incomment = true;
		}
		else if (!isspace(d_lines[line][column]))
		{
			return "";
		}
	}

	return "";
}

void
DocFinder::AddLocation(FileDescriptorProto const &proto,
                       Message const             &message,
                       deque<int> const          &path)
{
	SourceCodeInfo::Location location;

	if (FindPath(proto, path, location))
	{
		Documentation doc;
		string comment = ExtractDocumentation(location);

		if (!comment.empty())
		{
			doc.Location = location;
			doc.Comment = comment;

			d_locations[&message] = doc;
		}
	}
}

void
DocFinder::Scan(FileDescriptorProto const &proto,
                EnumDescriptorProto const &en,
                deque<int>                &path)
{
	AddLocation(proto, en, path);

	path.push_front(en.kValueFieldNumber);

	for (int i = 0; i < en.value_size(); ++i)
	{
		path.push_front(i);
		EnumValueDescriptorProto const &val = en.value(i);

		AddLocation(proto, val, path);

		path.pop_front();
	}

	path.pop_front();
}

void
DocFinder::Scan(FileDescriptorProto const  &proto,
                FieldDescriptorProto const &field,
                deque<int>                 &path)
{
	AddLocation(proto, field, path);
}

void
DocFinder::Scan(FileDescriptorProto const &proto,
                DescriptorProto     const &message,
                deque<int>                &path)
{
	AddLocation(proto, message, path);

	// Scan nested messages first
	path.push_front(message.kNestedTypeFieldNumber);

	for (int i = 0; i < message.nested_type_size(); ++i)
	{
		DescriptorProto const &nested = message.nested_type(i);

		path.push_front(i);
		Scan(proto, nested, path);
		path.pop_front();
	}

	path.pop_front();

	// Then scan enums
	path.push_front(message.kEnumTypeFieldNumber);

	for (int i = 0; i < message.enum_type_size(); ++i)
	{
		EnumDescriptorProto const &en = message.enum_type(i);

		path.push_front(i);
		Scan(proto, en, path);
		path.pop_front();
	}
	
	path.pop_front();

	// And finally fields
	path.push_front(message.kFieldFieldNumber);

	for (int i = 0; i < message.field_size(); ++i)
	{
		FieldDescriptorProto const &field = message.field(i);

		path.push_front(i);
		Scan(proto, field, path);
		path.pop_front();
	}

	path.pop_front();
}

void
DocFinder::Scan(FileDescriptorProto const &proto)
{
	ifstream fstr(d_filename.c_str());
	string line;

	while (getline(fstr, line))
	{
		line.erase(remove(line.begin(), line.end(), '\n'), line.end());
		line.erase(remove(line.begin(), line.end(), '\r'), line.end());

		size_t pos;

		while ((pos = line.find('\t')) != string::npos)
		{
			line.replace(pos, 1, "        ");
		}

		d_lines.push_back(line);
	}

	fstr.close();

	deque<int> path;
	path.push_front(proto.kMessageTypeFieldNumber);

	// Iterate over all messages
	for (int i = 0; i < proto.message_type_size(); ++i)
	{
		DescriptorProto const &message = proto.message_type(i);

		path.push_front(i);
		Scan(proto, message, path);
		path.pop_front();
	}

	path.pop_front();

	path.push_front(proto.kEnumTypeFieldNumber);

	// Iterate over all enums
	for (int i = 0; proto.enum_type_size(); ++i)
	{
		EnumDescriptorProto const &en = proto.enum_type(i);

		path.push_front(i);
		Scan(proto, en, path);
		path.pop_front();
	}
}
