#include "doxygen.hh"
#include "utils.hh"

#include <iostream>
#include <sstream>

using namespace libwebots::tools::protodoc;
using namespace std;

Doxygen::Doxygen(string const &description,
                 string const &extras)
:
	d_description(description),
	d_extras(extras)
{
}

Doxygen &
Doxygen::Brief(string const &description)
{
	d_brief = description;
	return *this;
}

Doxygen &
Doxygen::Param(string const &name,
               string const &description)
{
	d_params.push_back(pair<string, string>(name, description));
	return *this;
}

Doxygen &
Doxygen::Returns(string const &description)
{
	d_returns = description;
	return *this;
}

Doxygen &
Doxygen::Class(string const &name)
{
	d_className = name;
	return *this;
}

void
Doxygen::DoRemoveEmptyLines(vector<string> &lines,
                            vector<string>::iterator begin,
                            vector<string>::iterator end) const
{
	lines.erase(begin, end);
}

void
Doxygen::DoRemoveEmptyLines(vector<string> &lines,
                            vector<string>::reverse_iterator begin,
                            vector<string>::reverse_iterator end) const
{
	lines.erase(end.base(), begin.base());
}

string
Doxygen::NormalizeDescription() const
{
	if (d_description.empty() && d_extras.empty())
	{
		return "";
	}

	// Strip prefix "  *" from description
	vector<string> lines = Utils::Split(d_description, "\n");

	for (vector<string>::iterator iter = lines.begin(); iter != lines.end(); ++iter)
	{
		string stripped = Utils::StripLeft(Utils::StripLeft(*iter, " "), "*");

		if (Utils::StartsWith(stripped, " "))
		{
			stripped = stripped.substr(1);
		}

		*iter = stripped;
	}

	RemoveEmptyLines(lines, lines.begin(), lines.end());
	RemoveEmptyLines(lines, lines.rbegin(), lines.rend());

	if (!d_extras.empty())
	{
		lines.push_back("");
		lines.push_back("<em>Protobuf: [" + d_extras + "]</em>");
		lines.push_back("");
	}

	return " * " + Utils::Join(lines, "\n * ");
}

string
Doxygen::Generate() const
{
	string description = NormalizeDescription();

	if (description.empty() && d_brief.empty() && d_params.empty())
	{
		return "";
	}

	stringstream ret;

	ret << "/**" << endl;

	if (!d_className.empty())
	{
		ret << " * @class " << d_className << endl;
	}

	if (!d_brief.empty())
	{
		ret << " * @brief " << d_brief << endl;
	}

	vector<pair<string, string> >::const_iterator iter;

	for (iter = d_params.begin(); iter != d_params.end(); ++iter)
	{
		ret << " * @param " << iter->first << " " << iter->second << endl;
	}

	if (!d_brief.empty() || !d_params.empty())
	{
		ret << " *" << endl;
	}

	if (!description.empty())
	{
		ret << description << endl << " *" << endl;
	}

	if (!d_returns.empty())
	{
		ret << " * @returns " << d_returns << endl << " *" << endl;
	}

	ret << " **/";

	return ret.str();
}

ostream &
operator<<(ostream &stream, Doxygen const &doxygen)
{
	stream << doxygen.Generate();
	return stream;
}
