#include "generator.hh"
#include "indentstream.hh"
#include "doxygen.hh"
#include "utils.hh"

#include <iostream>
#include "commandline.hh"

using namespace google::protobuf;
using namespace libwebots::tools::protodoc;
using namespace std;

Generator::Generator(string const &filename,
                     FileDescriptorProto const &proto)
:
	d_filename(filename),
	d_proto(proto)
{
	d_docFinder = DocFinder(filename, proto);
}

void
Generator::Write(ostream &output)
{
	IndentStream stream(output);
	vector<string> ns;

	if (d_proto.has_package())
	{
		ns = SplitPackage(d_proto.package());

		vector<string>::iterator iter;

		for (iter = ns.begin(); iter != ns.end(); ++iter)
		{
			stream << "namespace " << *iter << endl << "{" << endl;
		}

		stream.Indent();
	}

	for (int i = 0; i < d_proto.message_type_size(); ++i)
	{
		Write(stream, d_proto.message_type(i));
		stream << endl;
	}

	for (int i = 0; i < d_proto.enum_type_size(); ++i)
	{
		Write(stream, d_proto.enum_type(i));
		stream << endl;
	}

	if (d_proto.has_package())
	{
		stream.Unindent();

		vector<string>::iterator iter;

		for (iter = ns.begin(); iter != ns.end(); ++iter)
		{
			stream << "}" << endl;
		}
	}
}

void
Generator::WriteComment(IndentStream  &output,
                        Message const &message)
{
	string comment;

	if (d_docFinder.Lookup(message, comment))
	{
		output << Doxygen(comment) << endl;
	}
}

string
Generator::ClassName(string const &name) const
{
	/* The file name, and then remove the prefix */
	CommandLine &cmd(CommandLine::Instance());

	string prefix = cmd.GetOption<CommandLine::OptionArg>("prefix").Argument;
	string output = cmd.GetOption<CommandLine::OptionArg>("output").Argument;

	string headername = Utils::Realpath(d_filename);

	if (!prefix.empty())
	{
		prefix = Utils::Realpath(prefix);
	}

	if (Utils::EndsWith(headername, ".proto"))
	{
		headername = headername.substr(0, headername.size() - 6) + ".pb.h";
	}

	if (output.empty() || output == "-")
	{
		return name;
	}

	if (!prefix.empty() && !Utils::EndsWith(prefix, "/"))
	{
		prefix += "/";
	}

	if (!prefix.empty() && Utils::StartsWith(headername, prefix))
	{
		return name + " " + Utils::Basename(output) + " " + headername.substr(prefix.size());
	}
	else
	{
		return name + " " + Utils::Basename(output) + " " + Utils::Basename(headername);
	}
}

void
Generator::WriteComment(IndentStream  &output,
                        DescriptorProto const &message)
{
	string comment;

	if (d_docFinder.Lookup(message, comment))
	{
		output << Doxygen(comment).Class(ClassName(message.name())) << endl;
	}
}

void
Generator::Write(IndentStream &output,
                 DescriptorProto const &message)
{
	WriteComment(output, message);

	output << "class " << message.name() << endl;
	output << "{" << endl;

	output.Indent();
	output << "public:" << endl;
	output.Indent();

	for (int i = 0; i < message.nested_type_size(); ++i)
	{
		Write(output, message.nested_type(i));
		output << endl;
	}

	for (int i = 0; i < message.enum_type_size(); ++i)
	{
		Write(output, message.enum_type(i));
		output << endl;
	}

	for (int i = 0; i < message.field_size(); ++i)
	{
		Write(output, message.field(i));
		output << endl;
	}

	output.Unindent();
	output.Unindent();

	output << "};" << endl;
}

void
Generator::Write(IndentStream &output,
                 EnumDescriptorProto const &en)
{
	WriteComment(output, en);

	output << "enum " << en.name() << endl;
	output << "{" << endl;

	output.Indent();

	for (int i = 0; i < en.value_size(); ++i)
	{
		Write(output, en.value(i));
	}

	output.Unindent();
	output << "};" << endl;
}

void
Generator::Write(IndentStream &output,
                 EnumValueDescriptorProto const &value)
{
	WriteComment(output, value);
	output << value.name() << " = " << value.number() << "," << endl;
}

void
Generator::Write(IndentStream &output,
                 FieldDescriptorProto const &field)
{
	string typestr;
	bool isprimitive = true;
	bool isstring = false;

	switch (field.type())
	{
		case FieldDescriptorProto::TYPE_GROUP :
		break;
		case FieldDescriptorProto::TYPE_DOUBLE:
			/* This is weird, but enums are somehow reported as
			   being doubles.. */
			typestr = field.type_name();

			if (typestr.empty())
			{
				typestr = "double";
			}
		break;
		case FieldDescriptorProto::TYPE_FLOAT:
			typestr = "float";
		break;
		case FieldDescriptorProto::TYPE_INT64:
		case FieldDescriptorProto::TYPE_SINT64:
		case FieldDescriptorProto::TYPE_SFIXED64:
			typestr = "::google::protobuf::int64";
		break;
		case FieldDescriptorProto::TYPE_UINT64:
		case FieldDescriptorProto::TYPE_FIXED64:
			typestr = "::google::protobuf::uint64";
		break;
		case FieldDescriptorProto::TYPE_INT32:
		case FieldDescriptorProto::TYPE_SINT32:
		case FieldDescriptorProto::TYPE_SFIXED32:
			typestr = "::google::protobuf::int32";
		break;
		case FieldDescriptorProto::TYPE_UINT32:
		case FieldDescriptorProto::TYPE_FIXED32:
			typestr = "::google::protobuf::uint32";
		break;
		case FieldDescriptorProto::TYPE_BOOL:
			typestr = "bool";
		break;
		case FieldDescriptorProto::TYPE_STRING:
		case FieldDescriptorProto::TYPE_BYTES:
			typestr = "std::string";
			isprimitive = false;
			isstring = true;
		break;
		case FieldDescriptorProto::TYPE_MESSAGE:
			isprimitive = false;
			typestr = field.type_name();
		break;
		case FieldDescriptorProto::TYPE_ENUM:
			/* Note that enums seem to be reported as doubles */
			isprimitive = false;
			typestr = field.type_name();
		break;
	}

	typestr = Utils::Replace(typestr, ".", "::");
	vector<string> extra;

	if (field.label() == FieldDescriptorProto::LABEL_OPTIONAL)
	{
		extra.push_back("optional");
	}
	else if (field.label() == FieldDescriptorProto::LABEL_REQUIRED)
	{
		extra.push_back("required");
	}

	if (field.has_default_value())
	{
		extra.push_back("default = " + field.default_value());
	}

	string extras = Utils::Join(extra, ", ");
	string desc = d_docFinder.Lookup(field);
	string nm = field.name();

	output << Doxygen(desc, extras).Brief("Clear the " + nm + " field.") << endl;
	output << "void clear_" << nm << "();" << endl << endl;

	if (field.label() != FieldDescriptorProto::LABEL_REPEATED)
	{
		output << Doxygen(desc, extras).
		          Brief("Check whether the @em " + nm + " field is set.").
		          Returns("true if the @em " + nm + " field is set, false otherwise.")
		       << endl;

		output << "bool has_" << nm << "() const;" << endl << endl;
	}
	else
	{
		output << Doxygen(desc, extras).
		          Brief("Get the number of @em " + nm + " elements.").
		          Returns("The number of @em " + nm + " elements.")
		       << endl;

		output << "int " << nm << "_size() const;" << endl << endl;
	}

	if (isprimitive)
	{
		if (field.label() == FieldDescriptorProto::LABEL_REPEATED)
		{
			output << Doxygen(desc, extras).
			          Brief("Get a @em " + nm + " element.").
			          Param("idx", "The element index.").
			          Returns("A @em " + nm + " element.")
			       << endl;

			output << typestr << " " << nm << "(int idx) const;" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Add a @em " + nm + " element.").
			          Param("value", "The element to add.")
			       << endl;

			output << "void add_" << nm << "(" << typestr << " value);" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Set a @em " + nm + " element.").
			          Param("idx", "The element index.").
			          Param("value", "The element to set.")
			       << endl;

			output << "void set_" << nm << "(int idx, " << typestr << " value);" << endl << endl;
		}
		else
		{
			output << Doxygen(desc, extras).
			          Brief("Get the @em " + nm + ".").
			          Returns("The @em " + nm + ".")
			       << endl;

			output << typestr << " " << nm << "() const;" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Set the @em " + nm + ".").
			          Param("value", "The @em " + nm + ".")
			       << endl;

			output << "void set_" << nm << "(" << typestr << " value);" << endl << endl;
		}
	}
	else
	{
		if (field.label() == FieldDescriptorProto::LABEL_REPEATED)
		{
			output << Doxygen(desc, extras).
			          Brief("Get a @em " + nm + " element.").
			          Param("idx", "The element index.").
			          Returns("A @em " + nm + " element.")
			       << endl;

			output << typestr << " const &" << nm << "(int idx) const;" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Get a mutable @em " + nm + " element.").
			          Param("idx", "The element index.").
			          Returns("A mutable @em " + nm + " element.")
			       << endl;

			output << typestr << " *mutable_" << nm << "(int idx);" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Add a new @em " + nm + " element.").
			          Returns("A new, mutable @em " + nm + " element.")
			       << endl;

			output << typestr << " *add_" << nm << "();" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Add a @em " + nm + " element.").
			          Param("value", "The element to add.")
			       << endl;

			output << "void add_" << nm << "(const " << typestr << " &value);" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Set a @em " + nm + " element.").
			          Param("idx", "The element index.").
			          Param("value", "The element to set.")
			       << endl;

			output << "void set_" << nm << "(int idx, const " << typestr << " &value);" << endl;

			if (isstring)
			{
				output << Doxygen(desc, extras).
				          Brief("Add a @em " + nm + " element, C string.").
				          Param("value", "The NULL terminated C string to add.")
				       << endl;

				output << "void add_" << nm << "(const char *value);" << endl << endl;

				output << Doxygen(desc, extras).
				          Brief("Add a @em " + nm + " element, C string with a size.").
				          Param("value", "The C string to add.").
				          Param("size", "The size of the string to add.")
				       << endl;

				output << "void add_" << nm << "(const char *value, size_t size);" << endl << endl;

				output << Doxygen(desc, extras).
				          Brief("Set a @em " + nm + " element, C string.").
				          Param("idx", "The element index.").
				          Param("value", "The NULL terminated C string to set.")
				       << endl;

				output << "void set_" << nm << "(int idx, const char *value);" << endl << endl;

				output << Doxygen(desc, extras).
				          Brief("Set a @em " + nm + " element, c string with a size.").
				          Param("idx", "The element index.").
				          Param("value", "The C string to set.").
				          Param("size", "The size of the string to set.")
				       << endl;

				output << "void set_" << nm << "(int idx, const char *value, size_t size);" << endl << endl;
			}
		}
		else
		{
			output << Doxygen(desc, extras).
			          Brief("Get the @em " + nm + ".").
			          Returns("The @em " + nm + ".")
			       << endl;

			output << "const " << typestr << " &" << nm << "() const;" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Get the mutable @em " + nm + ".").
			          Returns("The mutable @em " + nm + ".")
			       << endl;

			output << typestr << " *mutable_" << nm << "();" << endl << endl;

			output << Doxygen(desc, extras).
			          Brief("Set the @em " + nm + ".").
			          Param("value", "The @em " + nm + ".")
			       << endl;

			output << "void set_" << nm << "(const " << typestr << " &value);" << endl << endl;

			if (isstring)
			{
				output << Doxygen(desc, extras).
				          Brief("Set the @em " + nm + ", C string.").
				          Param("value", "A NULL terminated C string to set.")
				       << endl;

				output << "void set_" << nm << "(const char *value);" << endl << endl;

				output << Doxygen(desc, extras).
				          Brief("Set the @em " + nm + ", C string with a size.").
				          Param("value", "A C string to set.").
				          Param("size", "The size of the string to set.")
				       << endl;

				output << "void set_" << nm << "(const char *value, size_t size);" << endl << endl;


				output << Doxygen(desc, extras).
				          Brief("Release ownership the @em " + nm + " string.").
				          Returns("A pointer to the @em " + nm + " string.")
				       << endl;

				output << typestr << " *release_" << nm << "();" << endl << endl;
			}
		}
	}
}

vector<string>
Generator::SplitPackage(string const &package)
{
	size_t pos = 0;
	size_t nextpos = 0;

	vector<string> ret;

	while ((nextpos = package.find_first_of('.', pos)) != string::npos)
	{
		ret.push_back(package.substr(pos, nextpos - pos));
		pos = nextpos + 1;
	}

	if (package != "")
	{
		ret.push_back(package.substr(pos));
	}

	return ret;
}
