#include "protodoc.hh"
#include "commandline.hh"
#include "errorcollector.hh"
#include "generator.hh"
#include "utils.hh"

#include <google/protobuf/compiler/parser.h>
#include <google/protobuf/io/zero_copy_stream_impl.h>

#include <iostream>
#include <libgen.h>
#include <fstream>

using namespace google::protobuf::compiler;
using namespace google::protobuf;
using namespace google::protobuf::io;
using namespace libwebots::tools::protodoc;
using namespace std;

ProtoDoc::ProtoDoc(int argc, char * const *argv)
{
	CommandLine &cmd(CommandLine::Initialize("protodoc", "PROTOFILES"));

	cmd.Add(CommandLine::OptionArg("output", 'o', "Write output to specified file"));
	cmd.Add(CommandLine::OptionArg("prefix", 'p', "The include prefix"));

	CommandLine::Option &help = cmd.Add(CommandLine::Option("help",
	                                                        'h',
	                                                        "Show this help"));

	cmd.Parse(argc, argv);

	if (help)
	{
		cmd.ShowHelp();
		exit(1);
	}
}

bool
ProtoDoc::Generate(string const              &filename,
                   FileDescriptorProto const &fd,
                   ostream                   &output)
{
	Generator generator(filename, fd);

	generator.Write(output);

	return true;
}

bool
ProtoDoc::Generate(string const &filename,
                   ostream &output)
{
	ErrorCollector errorCollector;
	Parser parser;

	fstream fstr(filename.c_str(), ios::in);
	IstreamInputStream istr(&fstr);
	FileDescriptorProto fd;

	Tokenizer tokenizer(&istr, &errorCollector);

	if (!parser.Parse(&tokenizer, &fd))
	{
		cerr << "Failed to process `" << filename << "':" << endl;

		for (ErrorCollector::iterator it = errorCollector.begin();
		     it != errorCollector.end();
		     ++it)
		{
			cerr << "  "
			     << filename
			     << ":"
			     << it->Line
			     << "."
			     << it->Column
			     << ": "
			     << it->Message << endl;
		}

		return false;
	}
	else
	{
		return Generate(filename, fd, output);
	}
}

int
ProtoDoc::Run()
{
	CommandLine &cmd(CommandLine::Instance());
	vector<string> const &arguments = cmd.Arguments();

	if (arguments.size() == 0)
	{
		cerr << "Please provide some proto files" << endl;
		return 1;
	}

	ofstream fstr;
	string outputfile = cmd.GetOption<CommandLine::OptionArg>("output").Argument;
	bool usecout = (outputfile == "" || outputfile == "-");

	if (!usecout)
	{
		Utils::MakeDirs(Utils::Dirname(outputfile));
		fstr.open(outputfile.c_str());

		if (!fstr)
		{
			cerr << "Could not write to `" << outputfile << "'" << endl;
			return 1;
		}
	}

	ostream &output = usecout ? cout : fstr;

	for (vector<string>::const_iterator iter = arguments.begin();
	     iter != arguments.end();
	     ++iter)
	{
		if (!Generate(*iter, output))
		{
			return 1;
		}
	}

	return 0;
}

int
main(int argc, char * const argv[])
{
	ProtoDoc doc(argc, argv);
	return doc.Run();
}
