#include "utils.hh"

#include <sys/stat.h>
#include <cerrno>
#include <cstring>
#include <libgen.h>
#include <cstdlib>
#include <climits>

using namespace libwebots::tools::protodoc;
using namespace std;

vector<string>
Utils::Split(string const &s, string const &delim, size_t mmax, bool includeEmpty)
{
	vector<string> result;
	string::size_type pos = 0;

	while (mmax == 0 || result.size() < mmax)
	{
		string::size_type next = s.find(delim, pos);

		if (next == string::npos)
		{
			string part = s.substr(pos, s.size() - pos);

			if (part != "" || includeEmpty)
			{
				result.push_back(part);
			}

			pos = s.size();
			break;
		}
		else
		{
			string part = s.substr(pos, next - pos);

			if (part != "" || result.size() == 0 || includeEmpty)
			{
				result.push_back(part);
			}

			pos = next + delim.size();
		}
	}

	if (result.size() == 0)
	{
		result.push_back("");
	}

	if (pos != s.size())
	{
		result[result.size() - 1] += s.substr(pos - 1);
	}

	return result;
}

string
Utils::Join(vector<string> const &lines,
            string const &delimiter)
{
	string ret;

	for (vector<string>::const_iterator iter = lines.begin(); iter != lines.end(); ++iter)
	{
		if (iter != lines.begin())
		{
			ret += delimiter;
		}

		ret += *iter;
	}

	return ret;
}

string
Utils::StripLeft(string const &s,
                 string const &stripped)
{
	size_t pos = s.find_first_not_of(stripped);

	if (pos != string::npos)
	{
		return s.substr(pos);
	}
	else if (s.find_first_of(stripped) == 0)
	{
		return "";
	}
	else
	{
		return s;
	}
}

string
Utils::StripRight(string const &s,
                  string const &stripped)
{
	size_t pos = s.find_last_not_of(stripped);

	if (pos != string::npos && pos != 0)
	{
		return s.substr(0, pos + 1);
	}
	else if (s.find_first_not_of(stripped) != string::npos)
	{
		return "";
	}
	else
	{
		return s;
	}
}

string
Utils::Strip(string const &s,
             string const &stripped)
{
	return StripRight(StripLeft(s, stripped), stripped);
}

bool
Utils::StartsWith(string const &s,
                  string const &other)
{
	return s.substr(0, other.length()) == other;
}

bool
Utils::EndsWith(string const &s,
                string const &other)
{
	size_t l = s.length();
	size_t ol = other.length();

	if (l < ol)
	{
		return false;
	}

	return s.substr(l - ol) == other;
}

bool
Utils::MakeDirs(string const &dirs)
{
	vector<string> parts = Split(dirs, "/");
	string ret;

	for (vector<string>::iterator iter = parts.begin(); iter != parts.end(); ++iter)
	{
		if (iter != parts.begin())
		{
			ret += "/";
		}

		ret += *iter;

		if (!ret.empty())
		{
			int r = mkdir(ret.c_str(), 0755);

			if (r == -1 && errno != EEXIST)
			{
				return false;
			}
		}
	}

	return true;
}

string
Utils::Basename(string const &path)
{
	char *p = strdup(path.c_str());
	char *r = basename(p);

	if (!r)
	{
		return path;
	}

	string ret(r);
	free(p);

	return ret;
}

string
Utils::Dirname(string const &path)
{
	char *p = strdup(path.c_str());
	char *r = dirname(p);

	if (!r)
	{
		return path;
	}

	string ret(r);
	free(p);

	return ret;
}

string
Utils::Replace(string const &s,
               string const &find,
               string const &repl)
{
	string::size_type lastpos = 0;
	string ret = s;

	while (true)
	{
		string::size_type sp = ret.find(find, lastpos);

		if (sp == string::npos)
		{
			break;
		}

		ret.replace(sp, find.length(), repl);
		lastpos = sp + repl.length();
	}

	return ret;
}

string
Utils::Realpath(string const &path)
{
	char p[PATH_MAX];

	char *ptr = ::realpath(path.c_str(), p);

	if (!ptr)
	{
		return path;
	}

	return ptr;
}
